@@ -12,6 +12,8 @@ use std::sync::Arc;
 use std::time::{Duration, Instant};
 
 use emojis::Emoji;
+use matrix_sdk::ruma::events::receipt::ReceiptThread;
+use matrix_sdk::ruma::room_version_rules::RedactionRules;
 use ratatui::{
     buffer::Buffer,
     layout::{Alignment, Rect},
@@ -47,6 +49,7 @@ use matrix_sdk::{
             },
             room::redaction::{OriginalSyncRoomRedactionEvent, SyncRoomRedactionEvent},
             tag::{TagName, Tags},
+            AnySyncStateEvent,
             MessageLikeEvent,
         },
         presence::PresenceState,
@@ -55,7 +58,6 @@ use matrix_sdk::{
         OwnedRoomId,
         OwnedUserId,
         RoomId,
-        RoomVersionId,
         UserId,
     },
     RoomState as MatrixRoomState,
@@ -72,7 +74,7 @@ use modalkit::{
             ApplicationStore,
             ApplicationWindowId,
         },
-        completion::{complete_path, CompletionMap},
+        completion::{complete_path, Completer, CompletionMap},
         context::EditContext,
         cursor::Cursor,
         rope::EditRope,
@@ -90,6 +92,7 @@ use modalkit::{
 
 use crate::config::ImagePreviewProtocolValues;
 use crate::message::ImageStatus;
+use crate::notifications::NotificationHandle;
 use crate::preview::{source_from_event, spawn_insert_preview};
 use crate::{
     message::{Message, MessageEvent, MessageKey, MessageTimeStamp, Messages},
@@ -166,6 +169,9 @@ pub enum MessageAction {
     /// Reply to a message.
     Reply,
 
+    /// Go to the message the hovered message replied to.
+    Replied,
+
     /// Unreact to a message.
     ///
     /// If no specific Emoji to remove to is specified, then all reactions from the user on the
@@ -177,6 +183,19 @@ pub enum MessageAction {
     Unreact(Option<String>, bool),
 }
 
+/// An action taken in the currently selected space.
+#[derive(Clone, Debug, Eq, PartialEq)]
+pub enum SpaceAction {
+    /// Add a room or update metadata.
+    ///
+    /// The [`Option<String>`] argument is the order parameter.
+    /// The [`bool`] argument indicates whether the room is suggested.
+    SetChild(OwnedRoomId, Option<String>, bool),
+
+    /// Remove the selected room.
+    RemoveChild,
+}
+
 /// The type of room being created.
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub enum CreateRoomType {
@@ -243,6 +262,9 @@ pub enum SortFieldRoom {
 
     /// Sort rooms by the timestamps of their most recent messages.
     Recent,
+
+    /// Sort rooms by whether they are invites.
+    Invite,
 }
 
 /// Fields that users can be sorted by.
@@ -277,7 +299,7 @@ impl<'de> Deserialize<'de> for SortColumn<SortFieldRoom> {
 /// [serde] visitor for deserializing [SortColumn] for rooms and spaces.
 struct SortRoomVisitor;
 
-impl<'de> Visitor<'de> for SortRoomVisitor {
+impl Visitor<'_> for SortRoomVisitor {
     type Value = SortColumn<SortFieldRoom>;
 
     fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
@@ -307,6 +329,7 @@ impl<'de> Visitor<'de> for SortRoomVisitor {
             "name" => SortFieldRoom::Name,
             "alias" => SortFieldRoom::Alias,
             "id" => SortFieldRoom::RoomId,
+            "invite" => SortFieldRoom::Invite,
             _ => {
                 let msg = format!("Unknown sort field: {value:?}");
                 return Err(E::custom(msg));
@@ -329,7 +352,7 @@ impl<'de> Deserialize<'de> for SortColumn<SortFieldUser> {
 /// [serde] visitor for deserializing [SortColumn] for users.
 struct SortUserVisitor;
 
-impl<'de> Visitor<'de> for SortUserVisitor {
+impl Visitor<'_> for SortUserVisitor {
     type Value = SortColumn<SortFieldUser>;
 
     fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
@@ -375,6 +398,9 @@ pub enum RoomField {
     /// The room name.
     Name,
 
+    /// The room id.
+    Id,
+
     /// A room tag.
     Tag(TagName),
 
@@ -468,6 +494,8 @@ pub enum HomeserverAction {
     /// Create a new room with an optional localpart.
     CreateRoom(Option<String>, CreateRoomType, CreateRoomFlags),
     Logout(String, bool),
+    /// Forget all left rooms
+    Forget,
 }
 
 /// An action performed against the user's room keys.
@@ -493,6 +521,9 @@ pub enum IambAction {
     /// Perform an action on the currently selected message.
     Message(MessageAction),
 
+    /// Perform an action on the current space.
+    Space(SpaceAction),
+
     /// Open a URL.
     OpenLink(String),
 
@@ -534,6 +565,12 @@ impl From<MessageAction> for IambAction {
     }
 }
 
+impl From<SpaceAction> for IambAction {
+    fn from(act: SpaceAction) -> Self {
+        IambAction::Space(act)
+    }
+}
+
 impl From<RoomAction> for IambAction {
     fn from(act: RoomAction) -> Self {
         IambAction::Room(act)
@@ -553,6 +590,7 @@ impl ApplicationAction for IambAction {
             IambAction::Homeserver(..) => SequenceStatus::Break,
             IambAction::Keys(..) => SequenceStatus::Break,
             IambAction::Message(..) => SequenceStatus::Break,
+            IambAction::Space(..) => SequenceStatus::Break,
             IambAction::Room(..) => SequenceStatus::Break,
             IambAction::OpenLink(..) => SequenceStatus::Break,
             IambAction::Send(..) => SequenceStatus::Break,
@@ -568,6 +606,7 @@ impl ApplicationAction for IambAction {
             IambAction::Homeserver(..) => SequenceStatus::Atom,
             IambAction::Keys(..) => SequenceStatus::Atom,
             IambAction::Message(..) => SequenceStatus::Atom,
+            IambAction::Space(..) => SequenceStatus::Atom,
             IambAction::OpenLink(..) => SequenceStatus::Atom,
             IambAction::Room(..) => SequenceStatus::Atom,
             IambAction::Send(..) => SequenceStatus::Atom,
@@ -583,6 +622,7 @@ impl ApplicationAction for IambAction {
             IambAction::Homeserver(..) => SequenceStatus::Ignore,
             IambAction::Keys(..) => SequenceStatus::Ignore,
             IambAction::Message(..) => SequenceStatus::Ignore,
+            IambAction::Space(..) => SequenceStatus::Ignore,
             IambAction::Room(..) => SequenceStatus::Ignore,
             IambAction::OpenLink(..) => SequenceStatus::Ignore,
             IambAction::Send(..) => SequenceStatus::Ignore,
@@ -597,6 +637,7 @@ impl ApplicationAction for IambAction {
             IambAction::ClearUnreads => false,
             IambAction::Homeserver(..) => false,
             IambAction::Message(..) => false,
+            IambAction::Space(..) => false,
             IambAction::Room(..) => false,
             IambAction::Keys(..) => false,
             IambAction::Send(..) => false,
@@ -614,6 +655,12 @@ impl From<RoomAction> for ProgramAction {
     }
 }
 
+impl From<SpaceAction> for ProgramAction {
+    fn from(act: SpaceAction) -> Self {
+        IambAction::from(act).into()
+    }
+}
+
 impl From<IambAction> for ProgramAction {
     fn from(act: IambAction) -> Self {
         Action::Application(act)
@@ -709,10 +756,22 @@ pub enum IambError {
     #[error("Current window is not a room or space")]
     NoSelectedRoomOrSpace,
 
+    /// A failure due to not having a room or space item selected in a list.
+    #[error("No room or space currently selected in list")]
+    NoSelectedRoomOrSpaceItem,
+
     /// A failure due to not having a room selected.
     #[error("Current window is not a room")]
     NoSelectedRoom,
 
+    /// A failure due to not having a space selected.
+    #[error("Current window is not a space")]
+    NoSelectedSpace,
+
+    /// A failure due to not having sufficient permission to perform an action in a room.
+    #[error("You do not have the permission to do that")]
+    InsufficientPermission,
+
     /// A failure due to not having an outstanding room invitation.
     #[error("You do not have a current invitation to this room")]
     NotInvited,
@@ -729,6 +788,10 @@ pub enum IambError {
     #[error("Invalid room alias id: {0}")]
     InvalidRoomAliasId(#[from] matrix_sdk::ruma::IdParseError),
 
+    /// An invalid space child order was specified.
+    #[error("Invalid space child order: {0}")]
+    InvalidSpaceChildOrder(matrix_sdk::ruma::IdParseError),
+
     /// A failure occurred during verification.
     #[error("Verification request error: {0}")]
     VerificationRequestError(#[from] matrix_sdk::encryption::identities::RequestVerificationError),
@@ -785,6 +848,9 @@ pub enum EventLocation {
 
     /// The [EventId] belongs to a reaction to the given event.
     Reaction(OwnedEventId),
+
+    /// The [EventId] belongs to a state event in the main timeline of the room.
+    State(MessageKey),
 }
 
 impl EventLocation {
@@ -814,7 +880,6 @@ impl UnreadInfo {
 }
 
 /// Information about room's the user's joined.
-#[derive(Default)]
 pub struct RoomInfo {
     /// The display name for this room.
     pub name: Option<String>,
@@ -829,15 +894,13 @@ pub struct RoomInfo {
     messages: Messages,
 
     /// A map of read markers to display on different events.
-    pub event_receipts: HashMap<OwnedEventId, HashSet<OwnedUserId>>,
-
+    pub event_receipts: HashMap<ReceiptThread, HashMap<OwnedEventId, HashSet<OwnedUserId>>>,
     /// A map of the most recent read marker for each user.
     ///
-    /// Every receipt in this map should also have an entry in [`event_receipts`],
+    /// Every receipt in this map should also have an entry in [`event_receipts`](`Self::event_receipts`),
     /// however not every user has an entry. If a user's most recent receipt is
     /// older than the oldest loaded event, that user will not be included.
-    pub user_receipts: HashMap<OwnedUserId, OwnedEventId>,
-
+    pub user_receipts: HashMap<ReceiptThread, HashMap<OwnedUserId, OwnedEventId>>,
     /// A map of message identifiers to a map of reaction events.
     pub reactions: HashMap<OwnedEventId, MessageReactions>,
 
@@ -863,6 +926,28 @@ pub struct RoomInfo {
     pub draw_last: Option<Instant>,
 }
 
+impl Default for RoomInfo {
+    fn default() -> Self {
+        Self {
+            messages: Messages::new(ReceiptThread::Main),
+
+            name: Default::default(),
+            tags: Default::default(),
+            keys: Default::default(),
+            event_receipts: Default::default(),
+            user_receipts: Default::default(),
+            reactions: Default::default(),
+            threads: Default::default(),
+            fetching: Default::default(),
+            fetch_id: Default::default(),
+            fetch_last: Default::default(),
+            users_typing: Default::default(),
+            display_names: Default::default(),
+            draw_last: Default::default(),
+        }
+    }
+}
+
 impl RoomInfo {
     pub fn get_thread(&self, root: Option<&EventId>) -> Option<&Messages> {
         if let Some(thread_root) = root {
@@ -874,7 +959,9 @@ impl RoomInfo {
 
     pub fn get_thread_mut(&mut self, root: Option<OwnedEventId>) -> &mut Messages {
         if let Some(thread_root) = root {
-            self.threads.entry(thread_root).or_default()
+            self.threads
+                .entry(thread_root.clone())
+                .or_insert_with(|| Messages::thread(thread_root))
         } else {
             &mut self.messages
         }
@@ -945,24 +1032,30 @@ impl RoomInfo {
         self.messages.get_mut(self.keys.get(event_id)?.to_message_key()?)
     }
 
-    pub fn redact(&mut self, ev: OriginalSyncRoomRedactionEvent, room_version: &RoomVersionId) {
+    pub fn redact(&mut self, ev: OriginalSyncRoomRedactionEvent, rules: &RedactionRules) {
         let Some(redacts) = &ev.redacts else {
             return;
         };
 
         match self.keys.get(redacts) {
             None => return,
+            Some(EventLocation::State(key)) => {
+                if let Some(msg) = self.messages.get_mut(key) {
+                    let ev = SyncRoomRedactionEvent::Original(ev);
+                    msg.redact(ev, rules);
+                }
+            },
             Some(EventLocation::Message(None, key)) => {
                 if let Some(msg) = self.messages.get_mut(key) {
                     let ev = SyncRoomRedactionEvent::Original(ev);
-                    msg.redact(ev, room_version);
+                    msg.redact(ev, rules);
                 }
             },
             Some(EventLocation::Message(Some(root), key)) => {
                 if let Some(thread) = self.threads.get_mut(root) {
                     if let Some(msg) = thread.get_mut(key) {
                         let ev = SyncRoomRedactionEvent::Original(ev);
-                        msg.redact(ev, room_version);
+                        msg.redact(ev, rules);
                     }
                 }
             },
@@ -1008,7 +1101,9 @@ impl RoomInfo {
         };
 
         let source = if let Some(thread) = thread {
-            self.threads.entry(thread.clone()).or_default()
+            self.threads
+                .entry(thread.clone())
+                .or_insert_with(|| Messages::thread(thread.clone()))
         } else {
             &mut self.messages
         };
@@ -1025,6 +1120,7 @@ impl RoomInfo {
                 content.apply_replacement(new_msgtype);
             },
             MessageEvent::Redacted(_) |
+            MessageEvent::State(_) |
             MessageEvent::EncryptedOriginal(_) |
             MessageEvent::EncryptedRedacted(_) => {
                 return;
@@ -1034,16 +1130,52 @@ impl RoomInfo {
         msg.html = msg.event.html();
     }
 
+    pub fn insert_any_state(&mut self, msg: AnySyncStateEvent) {
+        let event_id = msg.event_id().to_owned();
+        let key = (msg.origin_server_ts().into(), event_id.clone());
+
+        let loc = EventLocation::State(key.clone());
+        self.keys.insert(event_id, loc);
+        self.messages.insert_message(key, msg);
+    }
+
     /// Indicates whether this room has unread messages.
     pub fn unreads(&self, settings: &ApplicationSettings) -> UnreadInfo {
         let last_message = self.messages.last_key_value();
-        let last_receipt = self.get_receipt(&settings.profile.user_id);
+
+        let last_receipt = self
+            .user_receipts
+            .get(&ReceiptThread::Main)
+            .and_then(|receipts| receipts.get(&settings.profile.user_id));
+        let last_receipt = last_receipt.as_ref().and_then(|event_id| {
+            match &self.keys.get(*event_id)? {
+                EventLocation::Message(_, key) | EventLocation::State(key) => Some(key),
+                EventLocation::Reaction(_) => None,
+            }
+        });
+
+        let last_unthreaded = self
+            .user_receipts
+            .get(&ReceiptThread::Unthreaded)
+            .and_then(|receipts| receipts.get(&settings.profile.user_id));
+        let last_unthreaded = last_unthreaded.as_ref().and_then(|event_id| {
+            match &self.keys.get(*event_id)? {
+                EventLocation::Message(_, key) | EventLocation::State(key) => Some(key),
+                EventLocation::Reaction(_) => None,
+            }
+        });
+
+        let last_receipt = std::cmp::max(last_receipt, last_unthreaded);
 
         match (last_message, last_receipt) {
-            (Some(((ts, recent), _)), Some(last_read)) => {
-                UnreadInfo { unread: last_read != recent, latest: Some(*ts) }
+            (Some(((ts, _), _)), Some((read_ts, _))) => {
+                UnreadInfo { unread: ts > read_ts, latest: Some(*ts) }
+            },
+            (Some(((ts, _), _)), None) => {
+                // If we've never loaded/generated a room's receipt (example,
+                // a newly joined but never viewed room), show it as unread.
+                UnreadInfo { unread: true, latest: Some(*ts) }
             },
-            (Some(((ts, _), _)), None) => UnreadInfo { unread: false, latest: Some(*ts) },
             (None, _) => UnreadInfo::default(),
         }
     }
@@ -1071,7 +1203,10 @@ impl RoomInfo {
         let event_id = msg.event_id().to_owned();
         let key = (msg.origin_server_ts().into(), event_id.clone());
 
-        let replies = self.threads.entry(thread_root.clone()).or_default();
+        let replies = self
+            .threads
+            .entry(thread_root.clone())
+            .or_insert_with(|| Messages::thread(thread_root.clone()));
         let loc = EventLocation::Message(Some(thread_root), key.clone());
         self.keys.insert(event_id, loc);
         replies.insert_message(key, msg);
@@ -1131,40 +1266,73 @@ impl RoomInfo {
 
     /// Indicates whether we've recently fetched scrollback for this room.
     pub fn recently_fetched(&self) -> bool {
-        self.fetch_last.map_or(false, |i| i.elapsed() < ROOM_FETCH_DEBOUNCE)
+        self.fetch_last.is_some_and(|i| i.elapsed() < ROOM_FETCH_DEBOUNCE)
     }
 
-    fn clear_receipt(&mut self, user_id: &OwnedUserId) -> Option<()> {
-        let old_event_id = self.user_receipts.get(user_id)?;
-        let old_receipts = self.event_receipts.get_mut(old_event_id)?;
+    fn clear_receipt(&mut self, thread: &ReceiptThread, user_id: &OwnedUserId) -> Option<()> {
+        let old_event_id =
+            self.user_receipts.get(thread).and_then(|receipts| receipts.get(user_id))?;
+        let old_thread = self.event_receipts.get_mut(thread)?;
+        let old_receipts = old_thread.get_mut(old_event_id)?;
         old_receipts.remove(user_id);
 
         if old_receipts.is_empty() {
-            self.event_receipts.remove(old_event_id);
+            old_thread.remove(old_event_id);
+        }
+        if old_thread.is_empty() {
+            self.event_receipts.remove(thread);
         }
 
         None
     }
 
-    pub fn set_receipt(&mut self, user_id: OwnedUserId, event_id: OwnedEventId) {
-        self.clear_receipt(&user_id);
+    pub fn set_receipt(
+        &mut self,
+        thread: ReceiptThread,
+        user_id: OwnedUserId,
+        event_id: OwnedEventId,
+    ) {
+        self.clear_receipt(&thread, &user_id);
         self.event_receipts
+            .entry(thread.clone())
+            .or_default()
             .entry(event_id.clone())
             .or_default()
             .insert(user_id.clone());
-        self.user_receipts.insert(user_id, event_id);
+        self.user_receipts.entry(thread).or_default().insert(user_id, event_id);
     }
 
-    pub fn fully_read(&mut self, user_id: OwnedUserId) {
+    pub fn fully_read(&mut self, user_id: &UserId) {
         let Some(((_, event_id), _)) = self.messages.last_key_value() else {
             return;
         };
 
-        self.set_receipt(user_id, event_id.clone());
+        self.set_receipt(ReceiptThread::Main, user_id.to_owned(), event_id.clone());
+
+        let newest = self
+            .threads
+            .iter()
+            .filter_map(|(thread_id, messages)| {
+                let thread = ReceiptThread::Thread(thread_id.to_owned());
+
+                messages
+                    .last_key_value()
+                    .map(|((_, event_id), _)| (thread, event_id.to_owned()))
+            })
+            .collect::<Vec<_>>();
+
+        for (thread, event_id) in newest.into_iter() {
+            self.set_receipt(thread, user_id.to_owned(), event_id.clone());
+        }
     }
 
-    pub fn get_receipt(&self, user_id: &UserId) -> Option<&OwnedEventId> {
-        self.user_receipts.get(user_id)
+    pub fn receipts<'a>(
+        &'a self,
+        user_id: &'a UserId,
+    ) -> impl Iterator<Item = (&'a ReceiptThread, &'a OwnedEventId)> + 'a {
+        self.user_receipts
+            .iter()
+            .filter_map(move |(t, rs)| rs.get(user_id).map(|r| (t, r)))
     }
 
     fn get_typers(&self) -> &[OwnedUserId] {
@@ -1223,7 +1391,9 @@ impl RoomInfo {
         }
 
         if !settings.tunables.typing_notice_display {
-            return area;
+            // still keep one line blank, so `render_jump_to_recent` doesn't immediately hide the
+            // last line in scrollback
+            return Rect::new(area.x, area.y, area.width, area.height - 1);
         }
 
         let top = Rect::new(area.x, area.y, area.width, area.height - 1);
@@ -1268,7 +1438,7 @@ fn emoji_map() -> CompletionMap<String, &'static Emoji> {
 
 #[cfg(unix)]
 fn picker_from_termios(protocol_type: Option<ProtocolType>) -> Option<Picker> {
-    let mut picker = match Picker::from_termios() {
+    let mut picker = match Picker::from_query_stdio() {
         Ok(picker) => picker,
         Err(e) => {
             tracing::error!("Failed to setup image previews: {e}");
@@ -1277,9 +1447,7 @@ fn picker_from_termios(protocol_type: Option<ProtocolType>) -> Option<Picker> {
     };
 
     if let Some(protocol_type) = protocol_type {
-        picker.protocol_type = protocol_type;
-    } else {
-        picker.guess_protocol();
+        picker.set_protocol_type(protocol_type);
     }
 
     Some(picker)
@@ -1302,8 +1470,8 @@ fn picker_from_settings(settings: &ApplicationSettings) -> Option<Picker> {
     }) = image_preview_protocol
     {
         // User forced type and font_size: use that.
-        let mut picker = Picker::new(font_size);
-        picker.protocol_type = protocol_type;
+        let mut picker = Picker::from_fontsize(font_size);
+        picker.set_protocol_type(protocol_type);
         Some(picker)
     } else {
         // Guess, but use type if forced.
@@ -1338,14 +1506,19 @@ impl SyncInfo {
     }
 }
 
-bitflags::bitflags! {
-    /// Load-needs
-    #[derive(Debug, Default, PartialEq)]
-    pub struct Need: u32 {
-        const EMPTY = 0b00000000;
-        const MESSAGES = 0b00000001;
-        const MEMBERS =  0b00000010;
-    }
+static MESSAGE_NEED_TTL: u8 = 30;
+
+#[derive(Debug, PartialEq)]
+/// Load messages until the event is loaded or `ttl` loads are exceeded
+pub struct MessageNeed {
+    pub event_id: OwnedEventId,
+    pub ttl: u8,
+}
+
+#[derive(Default, Debug, PartialEq)]
+pub struct Need {
+    pub members: bool,
+    pub messages: Option<Vec<MessageNeed>>,
 }
 
 /// Things that need loading for different rooms.
@@ -1355,9 +1528,31 @@ pub struct RoomNeeds {
 }
 
 impl RoomNeeds {
-    /// Mark a room for needing something to be loaded.
-    pub fn insert(&mut self, room_id: OwnedRoomId, need: Need) {
-        self.needs.entry(room_id).or_default().insert(need);
+    /// Mark a room for needing to load members.
+    pub fn need_members(&mut self, room_id: OwnedRoomId) {
+        self.needs.entry(room_id).or_default().members = true;
+    }
+
+    /// Mark a room for needing to load messages.
+    pub fn need_messages(&mut self, room_id: OwnedRoomId) {
+        self.needs.entry(room_id).or_default().messages.get_or_insert_default();
+    }
+
+    /// Mark a room for needing to load messages until the given message is loaded or a retry limit
+    /// is exceeded.
+    pub fn need_message(&mut self, room_id: OwnedRoomId, event_id: OwnedEventId) {
+        let messages = &mut self.needs.entry(room_id).or_default().messages.get_or_insert_default();
+
+        messages.push(MessageNeed { event_id, ttl: MESSAGE_NEED_TTL });
+    }
+
+    pub fn need_messages_all(&mut self, room_id: OwnedRoomId, message_needs: Vec<MessageNeed>) {
+        self.needs
+            .entry(room_id)
+            .or_default()
+            .messages
+            .get_or_insert_default()
+            .extend(message_needs);
     }
 
     pub fn rooms(&self) -> usize {
@@ -1417,6 +1612,12 @@ pub struct ChatStore {
 
     /// Whether the application is currently focused
     pub focused: bool,
+
+    /// Collator for locale-aware text sorting.
+    pub collator: feruca::Collator,
+
+    /// Notifications that should be dismissed when the user opens the room.
+    pub open_notifications: HashMap<OwnedRoomId, Vec<NotificationHandle>>,
 }
 
 impl ChatStore {
@@ -1431,6 +1632,7 @@ impl ChatStore {
             cmds: crate::commands::setup_commands(),
             emojis: emoji_map(),
 
+            collator: Default::default(),
             names: Default::default(),
             rooms: Default::default(),
             presences: Default::default(),
@@ -1440,6 +1642,7 @@ impl ChatStore {
             draw_curr: None,
             ring_bell: false,
             focused: true,
+            open_notifications: Default::default(),
         }
     }
 
@@ -1560,7 +1763,7 @@ impl<'de> Deserialize<'de> for IambId {
 /// [serde] visitor for deserializing [IambId].
 struct IambIdVisitor;
 
-impl<'de> Visitor<'de> for IambIdVisitor {
+impl Visitor<'_> for IambIdVisitor {
     type Value = IambId;
 
     fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
@@ -1697,6 +1900,13 @@ impl RoomFocus {
     pub fn is_msgbar(&self) -> bool {
         matches!(self, RoomFocus::MessageBar)
     }
+
+    pub fn toggle(&mut self) {
+        *self = match self {
+            RoomFocus::MessageBar => RoomFocus::Scrollback,
+            RoomFocus::Scrollback => RoomFocus::MessageBar,
+        };
+    }
 }
 
 /// Identifiers used to track where a mark was placed.
@@ -1765,11 +1975,20 @@ impl ApplicationInfo for IambInfo {
     type WindowId = IambId;
     type ContentId = IambBufferId;
 
+    fn content_of_command(ct: CommandType) -> IambBufferId {
+        IambBufferId::Command(ct)
+    }
+}
+
+pub struct IambCompleter;
+
+impl Completer<IambInfo> for IambCompleter {
     fn complete(
+        &mut self,
         text: &EditRope,
         cursor: &mut Cursor,
         content: &IambBufferId,
-        store: &mut ProgramStore,
+        store: &mut ChatStore,
     ) -> Vec<String> {
         match content {
             IambBufferId::Command(CommandType::Command) => complete_cmdbar(text, cursor, store),
@@ -1787,21 +2006,16 @@ impl ApplicationInfo for IambInfo {
             IambBufferId::UnreadList => vec![],
         }
     }
-
-    fn content_of_command(ct: CommandType) -> IambBufferId {
-        IambBufferId::Command(ct)
-    }
 }
 
 /// Tab completion for user IDs.
-fn complete_users(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -> Vec<String> {
+fn complete_users(text: &EditRope, cursor: &mut Cursor, store: &ChatStore) -> Vec<String> {
     let id = text
         .get_prefix_word_mut(cursor, &MATRIX_ID_WORD)
         .unwrap_or_else(EditRope::empty);
     let id = Cow::from(&id);
 
     store
-        .application
         .presences
         .complete(id.as_ref())
         .into_iter()
@@ -1810,7 +2024,7 @@ fn complete_users(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) ->
 }
 
 /// Tab completion within the message bar.
-fn complete_msgbar(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -> Vec<String> {
+fn complete_msgbar(text: &EditRope, cursor: &mut Cursor, store: &ChatStore) -> Vec<String> {
     let id = text
         .get_prefix_word_mut(cursor, &MATRIX_ID_WORD)
         .unwrap_or_else(EditRope::empty);
@@ -1819,13 +2033,12 @@ fn complete_msgbar(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -
     match id.chars().next() {
         // Complete room aliases.
         Some('#') => {
-            return store.application.names.complete(id.as_ref());
+            return store.names.complete(id.as_ref());
         },
 
         // Complete room identifiers.
         Some('!') => {
             return store
-                .application
                 .rooms
                 .complete(id.as_ref())
                 .into_iter()
@@ -1835,8 +2048,8 @@ fn complete_msgbar(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -
 
         // Complete Emoji shortcodes.
         Some(':') => {
-            let list = store.application.emojis.complete(&id[1..]);
-            let iter = list.into_iter().take(200).map(|s| format!(":{}:", s));
+            let list = store.emojis.complete(&id[1..]);
+            let iter = list.into_iter().take(200).map(|s| format!(":{s}:"));
 
             return iter.collect();
         },
@@ -1844,7 +2057,6 @@ fn complete_msgbar(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -
         // Complete usernames for @ and empty strings.
         Some('@') | None => {
             return store
-                .application
                 .presences
                 .complete(id.as_ref())
                 .into_iter()
@@ -1858,28 +2070,23 @@ fn complete_msgbar(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -
 }
 
 /// Tab completion for Matrix identifiers (usernames, room aliases, etc.)
-fn complete_matrix_names(
-    text: &EditRope,
-    cursor: &mut Cursor,
-    store: &ProgramStore,
-) -> Vec<String> {
+fn complete_matrix_names(text: &EditRope, cursor: &mut Cursor, store: &ChatStore) -> Vec<String> {
     let id = text
         .get_prefix_word_mut(cursor, &MATRIX_ID_WORD)
         .unwrap_or_else(EditRope::empty);
     let id = Cow::from(&id);
 
-    let list = store.application.names.complete(id.as_ref());
+    let list = store.names.complete(id.as_ref());
     if !list.is_empty() {
         return list;
     }
 
-    let list = store.application.presences.complete(id.as_ref());
+    let list = store.presences.complete(id.as_ref());
     if !list.is_empty() {
         return list.into_iter().map(|i| i.to_string()).collect();
     }
 
     store
-        .application
         .rooms
         .complete(id.as_ref())
         .into_iter()
@@ -1888,12 +2095,12 @@ fn complete_matrix_names(
 }
 
 /// Tab completion for Emoji shortcode names.
-fn complete_emoji(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -> Vec<String> {
+fn complete_emoji(text: &EditRope, cursor: &mut Cursor, store: &ChatStore) -> Vec<String> {
     let sc = text.get_prefix_word_mut(cursor, &WordStyle::Little);
     let sc = sc.unwrap_or_else(EditRope::empty);
     let sc = Cow::from(&sc);
 
-    store.application.emojis.complete(sc.as_ref())
+    store.emojis.complete(sc.as_ref())
 }
 
 /// Tab completion for command names.
@@ -1901,11 +2108,11 @@ fn complete_cmdname(
     desc: CommandDescription,
     text: &EditRope,
     cursor: &mut Cursor,
-    store: &ProgramStore,
+    store: &ChatStore,
 ) -> Vec<String> {
     // Complete command name and set cursor position.
     let _ = text.get_prefix_word_mut(cursor, &WordStyle::Little);
-    store.application.cmds.complete_name(desc.command.as_str())
+    store.cmds.complete_name(desc.command.as_str())
 }
 
 /// Tab completion for command arguments.
@@ -1913,9 +2120,9 @@ fn complete_cmdarg(
     desc: CommandDescription,
     text: &EditRope,
     cursor: &mut Cursor,
-    store: &ProgramStore,
+    store: &ChatStore,
 ) -> Vec<String> {
-    let cmd = match store.application.cmds.get(desc.command.as_str()) {
+    let cmd = match store.cmds.get(desc.command.as_str()) {
         Ok(cmd) => cmd,
         Err(_) => return vec![],
     };
@@ -1938,12 +2145,7 @@ fn complete_cmdarg(
 }
 
 /// Tab completion for commands.
-fn complete_cmd(
-    cmd: &str,
-    text: &EditRope,
-    cursor: &mut Cursor,
-    store: &ProgramStore,
-) -> Vec<String> {
+fn complete_cmd(cmd: &str, text: &EditRope, cursor: &mut Cursor, store: &ChatStore) -> Vec<String> {
     match CommandDescription::from_str(cmd) {
         Ok(desc) => {
             if desc.arg.untrimmed.is_empty() {
@@ -1960,7 +2162,7 @@ fn complete_cmd(
 }
 
 /// Tab completion for the command bar.
-fn complete_cmdbar(text: &EditRope, cursor: &mut Cursor, store: &ProgramStore) -> Vec<String> {
+fn complete_cmdbar(text: &EditRope, cursor: &mut Cursor, store: &ChatStore) -> Vec<String> {
     let eo = text.cursor_to_offset(cursor);
     let slice = text.slice(..eo);
     let cow = Cow::from(&slice);
@@ -1993,7 +2195,7 @@ pub mod tests {
         ));
 
         for i in 0..3 {
-            let event_id = format!("$house_{}", i);
+            let event_id = format!("$house_{i}");
             info.insert_reaction(MessageLikeEvent::Original(
                 matrix_sdk::ruma::events::OriginalMessageLikeEvent {
                     content: content.clone(),
@@ -2012,7 +2214,7 @@ pub mod tests {
         ));
 
         for i in 0..2 {
-            let event_id = format!("$smile_{}", i);
+            let event_id = format!("$smile_{i}");
             info.insert_reaction(MessageLikeEvent::Original(
                 matrix_sdk::ruma::events::OriginalMessageLikeEvent {
                     content: content.clone(),
@@ -2026,7 +2228,7 @@ pub mod tests {
         }
 
         for i in 2..4 {
-            let event_id = format!("$smile_{}", i);
+            let event_id = format!("$smile_{i}");
             info.insert_reaction(MessageLikeEvent::Original(
                 matrix_sdk::ruma::events::OriginalMessageLikeEvent {
                     content: content.clone(),
@@ -2128,18 +2330,19 @@ pub mod tests {
 
         let mut need_load = RoomNeeds::default();
 
-        need_load.insert(room_id.clone(), Need::MESSAGES);
-        need_load.insert(room_id.clone(), Need::MEMBERS);
+        need_load.need_messages(room_id.clone());
+        need_load.need_members(room_id.clone());
 
         assert_eq!(need_load.into_iter().collect::<Vec<(OwnedRoomId, Need)>>(), vec![(
             room_id,
-            Need::MESSAGES | Need::MEMBERS,
+            Need { members: true, messages: Some(Vec::new()) }
         )],);
     }
 
     #[tokio::test]
     async fn test_complete_msgbar() {
         let store = mock_store().await;
+        let store = store.application;
 
         let text = EditRope::from("going for a walk :walk ");
         let mut cursor = Cursor::new(0, 22);
@@ -2163,6 +2366,7 @@ pub mod tests {
     #[tokio::test]
     async fn test_complete_cmdbar() {
         let store = mock_store().await;
+        let store = store.application;
         let users = vec![
             "@user1:example.com",
             "@user2:example.com",
