@@ -1,16 +1,17 @@
 //! # Logic for loading and validating application configuration
 use std::borrow::Cow;
 use std::collections::hash_map::DefaultHasher;
-use std::collections::HashMap;
+use std::collections::{BTreeMap, HashMap};
+use std::env;
 use std::fmt;
 use std::fs::File;
 use std::hash::{Hash, Hasher};
-use std::io::{BufReader, BufWriter};
+use std::io::{BufReader, BufWriter, Write};
 use std::path::{Path, PathBuf};
 use std::process;
 
 use clap::Parser;
-use matrix_sdk::matrix_auth::MatrixSession;
+use matrix_sdk::authentication::matrix::MatrixSession;
 use matrix_sdk::ruma::{OwnedDeviceId, OwnedRoomAliasId, OwnedRoomId, OwnedUserId, UserId};
 use ratatui::style::{Color, Modifier as StyleModifier, Style};
 use ratatui::text::Span;
@@ -45,8 +46,9 @@ const DEFAULT_MEMBERS_SORT: [SortColumn<SortFieldUser>; 2] = [
     SortColumn(SortFieldUser::UserId, SortOrder::Ascending),
 ];
 
-const DEFAULT_ROOM_SORT: [SortColumn<SortFieldRoom>; 4] = [
+const DEFAULT_ROOM_SORT: [SortColumn<SortFieldRoom>; 5] = [
     SortColumn(SortFieldRoom::Favorite, SortOrder::Ascending),
+    SortColumn(SortFieldRoom::Invite, SortOrder::Ascending),
     SortColumn(SortFieldRoom::LowPriority, SortOrder::Ascending),
     SortColumn(SortFieldRoom::Unread, SortOrder::Ascending),
     SortColumn(SortFieldRoom::Name, SortOrder::Ascending),
@@ -97,14 +99,14 @@ fn validate_profile_name(name: &str) -> bool {
 
     let mut chars = name.chars();
 
-    if !chars.next().map_or(false, |c| c.is_ascii_alphanumeric()) {
+    if !chars.next().is_some_and(|c| c.is_ascii_alphanumeric()) {
         return false;
     }
 
     name.chars().all(is_profile_char)
 }
 
-fn validate_profile_names(names: &HashMap<String, ProfileConfig>) {
+fn validate_profile_names(names: &BTreeMap<String, ProfileConfig>) {
     for name in names.keys() {
         if validate_profile_name(name.as_str()) {
             continue;
@@ -151,7 +153,7 @@ pub enum ConfigError {
 pub struct Keys(pub Vec<TerminalKey>, pub String);
 pub struct KeysVisitor;
 
-impl<'de> Visitor<'de> for KeysVisitor {
+impl Visitor<'_> for KeysVisitor {
     type Value = Keys;
 
     fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
@@ -182,7 +184,7 @@ impl<'de> Deserialize<'de> for Keys {
 pub struct VimModes(pub Vec<VimMode>);
 pub struct VimModesVisitor;
 
-impl<'de> Visitor<'de> for VimModesVisitor {
+impl Visitor<'_> for VimModesVisitor {
     type Value = VimModes;
 
     fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
@@ -232,7 +234,7 @@ impl From<LogLevel> for Level {
     }
 }
 
-impl<'de> Visitor<'de> for LogLevelVisitor {
+impl Visitor<'_> for LogLevelVisitor {
     type Value = LogLevel;
 
     fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
@@ -267,7 +269,7 @@ impl<'de> Deserialize<'de> for LogLevel {
 pub struct UserColor(pub Color);
 pub struct UserColorVisitor;
 
-impl<'de> Visitor<'de> for UserColorVisitor {
+impl Visitor<'_> for UserColorVisitor {
     type Value = UserColor;
 
     fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
@@ -321,7 +323,7 @@ pub struct Session {
 impl From<Session> for MatrixSession {
     fn from(session: Session) -> Self {
         MatrixSession {
-            tokens: matrix_sdk::matrix_auth::MatrixSessionTokens {
+            tokens: matrix_sdk::authentication::SessionTokens {
                 access_token: session.access_token,
                 refresh_token: session.refresh_token,
             },
@@ -352,29 +354,31 @@ pub struct UserDisplayTunables {
 
 pub type UserOverrides = HashMap<OwnedUserId, UserDisplayTunables>;
 
-fn merge_sorts(a: SortOverrides, b: SortOverrides) -> SortOverrides {
+fn merge_sorts(profile: SortOverrides, global: SortOverrides) -> SortOverrides {
     SortOverrides {
-        chats: b.chats.or(a.chats),
-        dms: b.dms.or(a.dms),
-        rooms: b.rooms.or(a.rooms),
-        spaces: b.spaces.or(a.spaces),
-        members: b.members.or(a.members),
+        chats: profile.chats.or(global.chats),
+        dms: profile.dms.or(global.dms),
+        rooms: profile.rooms.or(global.rooms),
+        spaces: profile.spaces.or(global.spaces),
+        members: profile.members.or(global.members),
     }
 }
 
-fn merge_maps<K, V>(a: Option<HashMap<K, V>>, b: Option<HashMap<K, V>>) -> Option<HashMap<K, V>>
+fn merge_maps<K, V>(
+    profile: Option<HashMap<K, V>>,
+    global: Option<HashMap<K, V>>,
+) -> Option<HashMap<K, V>>
 where
     K: Eq + Hash,
 {
-    match (a, b) {
-        (Some(a), None) => Some(a),
-        (None, Some(b)) => Some(b),
-        (Some(mut a), Some(b)) => {
-            for (k, v) in b {
-                a.insert(k, v);
+    match (global, profile) {
+        (Some(m), None) | (None, Some(m)) => Some(m),
+        (Some(mut global), Some(profile)) => {
+            for (k, v) in profile {
+                global.insert(k, v);
             }
 
-            Some(a)
+            Some(global)
         },
         (None, None) => None,
     }
@@ -398,26 +402,79 @@ pub enum UserDisplayStyle {
     DisplayName,
 }
 
-#[derive(Clone, Copy, Debug, Deserialize, Eq, PartialEq)]
-#[serde(rename_all = "lowercase")]
-pub enum NotifyVia {
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+pub struct NotifyVia {
     /// Deliver notifications via terminal bell.
-    Bell,
+    pub bell: bool,
     /// Deliver notifications via desktop mechanism.
     #[cfg(feature = "desktop")]
-    Desktop,
+    pub desktop: bool,
 }
+pub struct NotifyViaVisitor;
 
 impl Default for NotifyVia {
     fn default() -> Self {
-        #[cfg(not(feature = "desktop"))]
-        return NotifyVia::Bell;
-
-        #[cfg(feature = "desktop")]
-        return NotifyVia::Desktop;
+        Self {
+            bell: cfg!(not(feature = "desktop")),
+            #[cfg(feature = "desktop")]
+            desktop: true,
+        }
     }
 }
 
+impl Visitor<'_> for NotifyViaVisitor {
+    type Value = NotifyVia;
+
+    fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
+        formatter.write_str("a valid notify destination (e.g. \"bell\" or \"desktop\")")
+    }
+
+    fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
+    where
+        E: SerdeError,
+    {
+        let mut via = NotifyVia {
+            bell: false,
+            #[cfg(feature = "desktop")]
+            desktop: false,
+        };
+
+        for value in value.split('|') {
+            match value.to_ascii_lowercase().as_str() {
+                "bell" => {
+                    via.bell = true;
+                },
+                #[cfg(feature = "desktop")]
+                "desktop" => {
+                    via.desktop = true;
+                },
+                #[cfg(not(feature = "desktop"))]
+                "desktop" => {
+                    return Err(E::custom("desktop notification support was compiled out"))
+                },
+                _ => return Err(E::custom("could not parse into a notify destination")),
+            };
+        }
+
+        Ok(via)
+    }
+}
+
+impl<'de> Deserialize<'de> for NotifyVia {
+    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+    where
+        D: Deserializer<'de>,
+    {
+        deserializer.deserialize_str(NotifyViaVisitor)
+    }
+}
+
+#[derive(Clone, Debug, Default, Deserialize, Eq, PartialEq)]
+pub struct Mouse {
+    #[serde(default)]
+    pub enabled: bool,
+}
+
 #[derive(Clone, Debug, Default, Deserialize, Eq, PartialEq)]
 pub struct Notifications {
     #[serde(default)]
@@ -426,6 +483,8 @@ pub struct Notifications {
     pub via: NotifyVia,
     #[serde(default = "default_true")]
     pub show_message: bool,
+    #[serde(default)]
+    pub sound_hint: Option<String>,
 }
 
 #[derive(Clone)]
@@ -501,12 +560,14 @@ impl SortOverrides {
 pub struct TunableValues {
     pub log_level: Level,
     pub message_shortcode_display: bool,
+    pub normal_after_send: bool,
     pub reaction_display: bool,
     pub reaction_shortcode_display: bool,
     pub read_receipt_send: bool,
     pub read_receipt_display: bool,
     pub request_timeout: u64,
     pub sort: SortValues,
+    pub state_event_display: bool,
     pub typing_notice_send: bool,
     pub typing_notice_display: bool,
     pub users: UserOverrides,
@@ -514,16 +575,19 @@ pub struct TunableValues {
     pub message_user_color: bool,
     pub default_room: Option<String>,
     pub open_command: Option<Vec<String>>,
+    pub mouse: Mouse,
     pub notifications: Notifications,
     pub image_preview: Option<ImagePreviewValues>,
     pub user_gutter_width: usize,
     pub external_edit_file_suffix: String,
+    pub tabstop: usize,
 }
 
 #[derive(Clone, Default, Deserialize)]
 pub struct Tunables {
     pub log_level: Option<LogLevel>,
     pub message_shortcode_display: Option<bool>,
+    pub normal_after_send: Option<bool>,
     pub reaction_display: Option<bool>,
     pub reaction_shortcode_display: Option<bool>,
     pub read_receipt_send: Option<bool>,
@@ -531,6 +595,7 @@ pub struct Tunables {
     pub request_timeout: Option<u64>,
     #[serde(default)]
     pub sort: SortOverrides,
+    pub state_event_display: Option<bool>,
     pub typing_notice_send: Option<bool>,
     pub typing_notice_display: Option<bool>,
     pub users: Option<UserOverrides>,
@@ -538,10 +603,12 @@ pub struct Tunables {
     pub message_user_color: Option<bool>,
     pub default_room: Option<String>,
     pub open_command: Option<Vec<String>>,
+    pub mouse: Option<Mouse>,
     pub notifications: Option<Notifications>,
     pub image_preview: Option<ImagePreview>,
     pub user_gutter_width: Option<usize>,
     pub external_edit_file_suffix: Option<String>,
+    pub tabstop: Option<usize>,
 }
 
 impl Tunables {
@@ -551,6 +618,7 @@ impl Tunables {
             message_shortcode_display: self
                 .message_shortcode_display
                 .or(other.message_shortcode_display),
+            normal_after_send: self.normal_after_send.or(other.normal_after_send),
             reaction_display: self.reaction_display.or(other.reaction_display),
             reaction_shortcode_display: self
                 .reaction_shortcode_display
@@ -559,6 +627,7 @@ impl Tunables {
             read_receipt_display: self.read_receipt_display.or(other.read_receipt_display),
             request_timeout: self.request_timeout.or(other.request_timeout),
             sort: merge_sorts(self.sort, other.sort),
+            state_event_display: self.state_event_display.or(other.state_event_display),
             typing_notice_send: self.typing_notice_send.or(other.typing_notice_send),
             typing_notice_display: self.typing_notice_display.or(other.typing_notice_display),
             users: merge_maps(self.users, other.users),
@@ -566,12 +635,14 @@ impl Tunables {
             message_user_color: self.message_user_color.or(other.message_user_color),
             default_room: self.default_room.or(other.default_room),
             open_command: self.open_command.or(other.open_command),
+            mouse: self.mouse.or(other.mouse),
             notifications: self.notifications.or(other.notifications),
             image_preview: self.image_preview.or(other.image_preview),
             user_gutter_width: self.user_gutter_width.or(other.user_gutter_width),
             external_edit_file_suffix: self
                 .external_edit_file_suffix
                 .or(other.external_edit_file_suffix),
+            tabstop: self.tabstop.or(other.tabstop),
         }
     }
 
@@ -579,12 +650,14 @@ impl Tunables {
         TunableValues {
             log_level: self.log_level.map(Level::from).unwrap_or(Level::INFO),
             message_shortcode_display: self.message_shortcode_display.unwrap_or(false),
+            normal_after_send: self.normal_after_send.unwrap_or(false),
             reaction_display: self.reaction_display.unwrap_or(true),
             reaction_shortcode_display: self.reaction_shortcode_display.unwrap_or(false),
             read_receipt_send: self.read_receipt_send.unwrap_or(true),
             read_receipt_display: self.read_receipt_display.unwrap_or(true),
             request_timeout: self.request_timeout.unwrap_or(DEFAULT_REQ_TIMEOUT),
             sort: self.sort.values(),
+            state_event_display: self.state_event_display.unwrap_or(true),
             typing_notice_send: self.typing_notice_send.unwrap_or(true),
             typing_notice_display: self.typing_notice_display.unwrap_or(true),
             users: self.users.unwrap_or_default(),
@@ -592,12 +665,14 @@ impl Tunables {
             message_user_color: self.message_user_color.unwrap_or(false),
             default_room: self.default_room,
             open_command: self.open_command,
+            mouse: self.mouse.unwrap_or_default(),
             notifications: self.notifications.unwrap_or_default(),
             image_preview: self.image_preview.map(ImagePreview::values),
             user_gutter_width: self.user_gutter_width.unwrap_or(30),
             external_edit_file_suffix: self
                 .external_edit_file_suffix
                 .unwrap_or_else(|| ".md".to_string()),
+            tabstop: self.tabstop.unwrap_or(4),
         }
     }
 }
@@ -632,11 +707,11 @@ impl DirectoryValues {
 
 #[derive(Clone, Default, Deserialize)]
 pub struct Directories {
-    pub cache: Option<PathBuf>,
-    pub data: Option<PathBuf>,
-    pub logs: Option<PathBuf>,
-    pub downloads: Option<PathBuf>,
-    pub image_previews: Option<PathBuf>,
+    pub cache: Option<String>,
+    pub data: Option<String>,
+    pub logs: Option<String>,
+    pub downloads: Option<String>,
+    pub image_previews: Option<String>,
 }
 
 impl Directories {
@@ -653,6 +728,11 @@ impl Directories {
     fn values(self) -> DirectoryValues {
         let cache = self
             .cache
+            .map(|dir| {
+                let dir = shellexpand::full(&dir)
+                    .expect("unable to expand shell variables in dirs.cache");
+                Path::new(dir.as_ref()).to_owned()
+            })
             .or_else(|| {
                 let mut dir = dirs::cache_dir()?;
                 dir.push("iamb");
@@ -662,6 +742,11 @@ impl Directories {
 
         let data = self
             .data
+            .map(|dir| {
+                let dir = shellexpand::full(&dir)
+                    .expect("unable to expand shell variables in dirs.cache");
+                Path::new(dir.as_ref()).to_owned()
+            })
             .or_else(|| {
                 let mut dir = dirs::data_dir()?;
                 dir.push("iamb");
@@ -669,19 +754,40 @@ impl Directories {
             })
             .expect("no dirs.data value configured!");
 
-        let logs = self.logs.unwrap_or_else(|| {
-            let mut dir = cache.clone();
-            dir.push("logs");
-            dir
-        });
+        let logs = self
+            .logs
+            .map(|dir| {
+                let dir = shellexpand::full(&dir)
+                    .expect("unable to expand shell variables in dirs.cache");
+                Path::new(dir.as_ref()).to_owned()
+            })
+            .unwrap_or_else(|| {
+                let mut dir = cache.clone();
+                dir.push("logs");
+                dir
+            });
 
-        let downloads = self.downloads.or_else(dirs::download_dir);
+        let downloads = self
+            .downloads
+            .map(|dir| {
+                let dir = shellexpand::full(&dir)
+                    .expect("unable to expand shell variables in dirs.cache");
+                Path::new(dir.as_ref()).to_owned()
+            })
+            .or_else(dirs::download_dir);
 
-        let image_previews = self.image_previews.unwrap_or_else(|| {
-            let mut dir = cache.clone();
-            dir.push("image_preview_downloads");
-            dir
-        });
+        let image_previews = self
+            .image_previews
+            .map(|dir| {
+                let dir = shellexpand::full(&dir)
+                    .expect("unable to expand shell variables in dirs.cache");
+                Path::new(dir.as_ref()).to_owned()
+            })
+            .unwrap_or_else(|| {
+                let mut dir = cache.clone();
+                dir.push("image_preview_downloads");
+                dir
+            });
 
         DirectoryValues { cache, data, logs, downloads, image_previews }
     }
@@ -729,7 +835,7 @@ pub struct ProfileConfig {
 
 #[derive(Clone, Deserialize)]
 pub struct IambConfig {
-    pub profiles: HashMap<String, ProfileConfig>,
+    pub profiles: BTreeMap<String, ProfileConfig>,
     pub default_profile: Option<String>,
     pub settings: Option<Tunables>,
     pub dirs: Option<Directories>,
@@ -769,14 +875,22 @@ pub struct ApplicationSettings {
 }
 
 impl ApplicationSettings {
+    fn get_xdg_config_home() -> Option<PathBuf> {
+        env::var("XDG_CONFIG_HOME").ok().map(PathBuf::from)
+    }
+
     pub fn load(cli: Iamb) -> Result<Self, Box<dyn std::error::Error>> {
-        let mut config_dir = cli.config_directory.or_else(dirs::config_dir).unwrap_or_else(|| {
-            usage!(
-                "No user configuration directory found;\
-                please specify one via -C.\n\n
-                For more information try '--help'"
-            );
-        });
+        let mut config_dir = cli
+            .config_directory
+            .or_else(Self::get_xdg_config_home)
+            .or_else(dirs::config_dir)
+            .unwrap_or_else(|| {
+                usage!(
+                    "No user configuration directory found;\
+                    please specify one via -C.\n\n
+                    For more information try '--help'"
+                );
+            });
 
         config_dir.push("iamb");
         let config_json = config_dir.join("config.json");
@@ -816,14 +930,33 @@ impl ApplicationSettings {
         } else if profiles.len() == 1 {
             profiles.into_iter().next().unwrap()
         } else {
-            usage!(
-                "No profile specified. \
-                Please use -P or add \"default_profile\" to your configuration.\n\n\
-                For more information try '--help'",
-            );
+            loop {
+                println!("\nNo profile specified. Available profiles:");
+                profiles.keys().enumerate().for_each(|(i, name)| println!("{i}: {name}"));
+
+                print!("Select a number or 'q' to quit: ");
+                let _ = std::io::stdout().flush();
+
+                let mut input = String::new();
+                let _ = std::io::stdin().read_line(&mut input);
+
+                if input.trim() == "q" {
+                    usage!(
+                        "No profile specified. \
+                        Please use -P or add \"default_profile\" to your configuration.\n\n\
+                        For more information try '--help'",
+                    );
+                }
+                if let Ok(i) = input.trim().parse::<usize>() {
+                    if i < profiles.len() {
+                        break profiles.into_iter().nth(i).unwrap();
+                    }
+                }
+                println!("\nInvalid index.");
+            }
         };
 
-        let macros = merge_maps(macros, profile.macros.take()).unwrap_or_default();
+        let macros = merge_maps(profile.macros.take(), macros).unwrap_or_default();
         let layout = profile.layout.take().or(layout).unwrap_or_default();
 
         let tunables = global.unwrap_or_default();
@@ -898,7 +1031,7 @@ impl ApplicationSettings {
         Ok(())
     }
 
-    pub fn get_user_char_span<'a>(&self, user_id: &'a UserId) -> Span<'a> {
+    pub fn get_user_char_span(&self, user_id: &UserId) -> Span<'_> {
         let (color, c) = self
             .tunables
             .users
@@ -1022,10 +1155,10 @@ mod tests {
         assert_eq!(res, Some(b.clone()));
 
         let res = merge_maps(Some(b.clone()), Some(c.clone()));
-        assert_eq!(res, Some(c.clone()));
+        assert_eq!(res, Some(b.clone()));
 
         let res = merge_maps(Some(c.clone()), Some(b.clone()));
-        assert_eq!(res, Some(b.clone()));
+        assert_eq!(res, Some(c.clone()));
     }
 
     #[test]
@@ -1189,6 +1322,29 @@ mod tests {
         assert_eq!(run, &exp);
     }
 
+    #[test]
+    fn test_parse_notify_via() {
+        assert_eq!(NotifyVia { bell: false, desktop: true }, NotifyVia::default());
+        assert_eq!(
+            NotifyVia { bell: false, desktop: true },
+            serde_json::from_str(r#""desktop""#).unwrap()
+        );
+        assert_eq!(
+            NotifyVia { bell: true, desktop: false },
+            serde_json::from_str(r#""bell""#).unwrap()
+        );
+        assert_eq!(
+            NotifyVia { bell: true, desktop: true },
+            serde_json::from_str(r#""bell|desktop""#).unwrap()
+        );
+        assert_eq!(
+            NotifyVia { bell: true, desktop: true },
+            serde_json::from_str(r#""desktop|bell""#).unwrap()
+        );
+        assert!(serde_json::from_str::<NotifyVia>(r#""other""#).is_err());
+        assert!(serde_json::from_str::<NotifyVia>(r#""""#).is_err());
+    }
+
     #[test]
     fn test_load_example_config_toml() {
         let path = PathBuf::from("config.example.toml");
