@@ -44,11 +44,14 @@ use modalkit::crossterm::{
         read,
         DisableBracketedPaste,
         DisableFocusChange,
+        DisableMouseCapture,
         EnableBracketedPaste,
         EnableFocusChange,
+        EnableMouseCapture,
         Event,
         KeyEventKind,
         KeyboardEnhancementFlags,
+        MouseEventKind,
         PopKeyboardEnhancementFlags,
         PushKeyboardEnhancementFlags,
     },
@@ -59,7 +62,7 @@ use modalkit::crossterm::{
 use ratatui::{
     backend::CrosstermBackend,
     layout::Rect,
-    style::{Color, Style},
+    style::{Color, Modifier, Style},
     text::Span,
     widgets::Paragraph,
     Terminal,
@@ -86,6 +89,7 @@ use crate::{
         ChatStore,
         HomeserverAction,
         IambAction,
+        IambCompleter,
         IambError,
         IambId,
         IambInfo,
@@ -310,7 +314,7 @@ impl Application {
         }
 
         term.draw(|f| {
-            let area = f.size();
+            let area = f.area();
 
             let modestr = bindings.show_mode();
             let cursor = bindings.get_cursor_indicator();
@@ -324,6 +328,9 @@ impl Application {
                 .show_dialog(dialogstr)
                 .show_mode(modestr)
                 .borders(true)
+                .border_style(Style::default().add_modifier(Modifier::DIM))
+                .tab_style(Style::default().add_modifier(Modifier::DIM))
+                .tab_style_focused(Style::default().remove_modifier(Modifier::DIM))
                 .focus(focused);
             f.render_stateful_widget(screen, area, sstate);
 
@@ -339,7 +346,7 @@ impl Application {
                     let inner = Rect::new(cx, cy, 1, 1);
                     f.render_widget(para, inner)
                 }
-                f.set_cursor(cx, cy);
+                f.set_cursor_position((cx, cy));
             }
         })?;
 
@@ -364,8 +371,30 @@ impl Application {
 
                     return Ok(ke.into());
                 },
-                Event::Mouse(_) => {
-                    // Do nothing for now.
+                Event::Mouse(me) => {
+                    let dir = match me.kind {
+                        MouseEventKind::ScrollUp => MoveDir2D::Up,
+                        MouseEventKind::ScrollDown => MoveDir2D::Down,
+                        MouseEventKind::ScrollLeft => MoveDir2D::Left,
+                        MouseEventKind::ScrollRight => MoveDir2D::Right,
+                        _ => continue,
+                    };
+
+                    let size = ScrollSize::Cell;
+                    let style = ScrollStyle::Direction2D(dir, size, 1.into());
+                    let ctx = ProgramContext::default();
+                    let mut store = self.store.lock().await;
+
+                    match self.screen.scroll(&style, &ctx, store.deref_mut()) {
+                        Ok(None) => {},
+                        Ok(Some(info)) => {
+                            drop(store);
+                            self.handle_info(info);
+                        },
+                        Err(e) => {
+                            self.screen.push_error(e);
+                        },
+                    }
                 },
                 Event::FocusGained => {
                     let mut store = self.store.lock().await;
@@ -504,7 +533,7 @@ impl Application {
             },
 
             // Unimplemented.
-            Action::KeywordLookup => {
+            Action::KeywordLookup(_) => {
                 // XXX: implement
                 None
             },
@@ -532,9 +561,12 @@ impl Application {
             IambAction::ClearUnreads => {
                 let user_id = &store.application.settings.profile.user_id;
 
+                // Clear any notifications we displayed:
+                store.application.open_notifications.clear();
+
                 for room_id in store.application.sync_info.chats() {
                     if let Some(room) = store.application.rooms.get_mut(room_id) {
-                        room.fully_read(user_id.clone());
+                        room.fully_read(user_id);
                     }
                 }
 
@@ -557,6 +589,9 @@ impl Application {
             IambAction::Message(act) => {
                 self.screen.current_window_mut()?.message_command(act, ctx, store).await?
             },
+            IambAction::Space(act) => {
+                self.screen.current_window_mut()?.space_command(act, ctx, store).await?
+            },
             IambAction::Room(act) => {
                 let acts = self.screen.current_window_mut()?.room_command(act, ctx, store).await?;
                 self.action_prepend(acts);
@@ -564,6 +599,9 @@ impl Application {
                 None
             },
             IambAction::Send(act) => {
+                if store.application.settings.tunables.normal_after_send {
+                    self.bindings.reset_mode();
+                }
                 self.screen.current_window_mut()?.send_command(act, ctx, store).await?
             },
 
@@ -625,6 +663,13 @@ impl Application {
 
                 Err(UIError::NeedConfirm(prompt))
             },
+            HomeserverAction::Forget => {
+                let client = &store.application.worker.client;
+                for room in client.left_rooms() {
+                    room.forget().await.map_err(IambError::from)?;
+                }
+                Ok(vec![])
+            },
         }
     }
 
@@ -847,7 +892,7 @@ async fn check_import_keys(
     let encrypted = match encrypt_room_key_export(&keys, &passphrase, 500000) {
         Ok(encrypted) => encrypted,
         Err(e) => {
-            format!("* Failed to encrypt room keys during export: {e}");
+            println!("* Failed to encrypt room keys during export: {e}");
             process::exit(2);
         },
     };
@@ -929,8 +974,8 @@ async fn login_normal(
 }
 
 /// Set up the terminal for drawing the TUI, and getting additional info.
-fn setup_tty(title: &str, enable_enhanced_keys: bool) -> std::io::Result<()> {
-    let title = format!("iamb ({})", title);
+fn setup_tty(settings: &ApplicationSettings, enable_enhanced_keys: bool) -> std::io::Result<()> {
+    let title = format!("iamb ({})", settings.profile.user_id.as_str());
 
     // Enable raw mode and enter the alternate screen.
     crossterm::terminal::enable_raw_mode()?;
@@ -944,15 +989,23 @@ fn setup_tty(title: &str, enable_enhanced_keys: bool) -> std::io::Result<()> {
         )?;
     }
 
+    if settings.tunables.mouse.enabled {
+        crossterm::execute!(stdout(), EnableMouseCapture)?;
+    }
+
     crossterm::execute!(stdout(), EnableBracketedPaste, EnableFocusChange, SetTitle(title))
 }
 
 // Do our best to reverse what we did in setup_tty() when we exit or crash.
-fn restore_tty(enable_enhanced_keys: bool) {
+fn restore_tty(enable_enhanced_keys: bool, enable_mouse: bool) {
     if enable_enhanced_keys {
         let _ = crossterm::queue!(stdout(), PopKeyboardEnhancementFlags);
     }
 
+    if enable_mouse {
+        let _ = crossterm::queue!(stdout(), DisableMouseCapture);
+    }
+
     let _ = crossterm::execute!(
         stdout(),
         DisableBracketedPaste,
@@ -975,7 +1028,9 @@ async fn run(settings: ApplicationSettings) -> IambResult<()> {
     // Set up the async worker thread and global store.
     let worker = ClientWorker::spawn(client.clone(), settings.clone()).await;
     let store = ChatStore::new(worker.clone(), settings.clone());
-    let store = Store::new(store);
+    let mut store = Store::new(store);
+    store.completer = Box::new(IambCompleter);
+
     let store = Arc::new(AsyncMutex::new(store));
     worker.init(store.clone());
 
@@ -988,7 +1043,7 @@ async fn run(settings: ApplicationSettings) -> IambResult<()> {
     match res {
         Err(UIError::Application(IambError::Matrix(e))) => {
             if let Some(ErrorKind::UnknownToken { .. }) = e.client_api_error_kind() {
-                print_exit("Server did not recognize our API token; did you log out from this session elsewhere?")
+                print_exit(format!("Server did not recognize our API token; did you log out from this session elsewhere?\nTry deleting `{}` to force a clean login.", settings.session_json.display()))
             } else {
                 print_exit(e)
             }
@@ -1006,11 +1061,12 @@ async fn run(settings: ApplicationSettings) -> IambResult<()> {
             false
         },
     };
-    setup_tty(settings.profile.user_id.as_str(), enable_enhanced_keys)?;
+    setup_tty(&settings, enable_enhanced_keys)?;
 
     let orig_hook = std::panic::take_hook();
+    let enable_mouse = settings.tunables.mouse.enabled;
     std::panic::set_hook(Box::new(move |panic_info| {
-        restore_tty(enable_enhanced_keys);
+        restore_tty(enable_enhanced_keys, enable_mouse);
         orig_hook(panic_info);
         process::exit(1);
     }));
@@ -1020,7 +1076,7 @@ async fn run(settings: ApplicationSettings) -> IambResult<()> {
     application.run().await?;
 
     // Clean up the terminal on exit.
-    restore_tty(enable_enhanced_keys);
+    restore_tty(enable_enhanced_keys, enable_mouse);
 
     Ok(())
 }
