@@ -10,10 +10,12 @@
 //!
 //! This isn't as important for iamb, since it isn't a browser environment, but we do still map
 //! input onto an enum of the safe list of tags to keep it easy to understand and process.
+use std::borrow::Cow;
 use std::ops::Deref;
 
 use css_color_parser::Color as CssColor;
 use markup5ever_rcdom::{Handle, NodeData, RcDom};
+use matrix_sdk::ruma::{OwnedRoomAliasId, OwnedRoomId, OwnedUserId};
 use unicode_segmentation::UnicodeSegmentation;
 use url::Url;
 
@@ -34,10 +36,13 @@ use ratatui::{
 };
 
 use crate::{
+    config::ApplicationSettings,
     message::printer::TextPrinter,
     util::{join_cell_text, space_text},
 };
 
+const QUOTE_COLOR: Color = Color::Indexed(236);
+
 /// Generate bullet points from a [ListStyle].
 pub struct BulletIterator {
     style: ListStyle,
@@ -148,7 +153,12 @@ impl Table {
         }
     }
 
-    fn to_text(&self, width: usize, style: Style, emoji_shortcodes: bool) -> Text {
+    fn to_text<'a>(
+        &'a self,
+        width: usize,
+        style: Style,
+        settings: &'a ApplicationSettings,
+    ) -> Text<'a> {
         let mut text = Text::default();
         let columns = self.columns();
         let cell_total = width.saturating_sub(columns).saturating_sub(1);
@@ -167,7 +177,7 @@ impl Table {
         if let Some(caption) = &self.caption {
             let subw = width.saturating_sub(6);
             let mut printer =
-                TextPrinter::new(subw, style, true, emoji_shortcodes).align(Alignment::Center);
+                TextPrinter::new(subw, style, true, settings).align(Alignment::Center);
             caption.print(&mut printer, style);
 
             for mut line in printer.finish().lines {
@@ -214,7 +224,7 @@ impl Table {
                                 CellType::Data => style,
                             };
 
-                            cell.to_text(*w, style, emoji_shortcodes)
+                            cell.to_text(*w, style, settings)
                         } else {
                             space_text(*w, style)
                         };
@@ -271,13 +281,22 @@ pub enum StyleTreeNode {
     Ruler,
     Style(Box<StyleTreeNode>, Style),
     Table(Table),
-    Text(String),
+    Text(Cow<'static, str>),
     Sequence(StyleTreeChildren),
+    RoomAlias(OwnedRoomAliasId),
+    RoomId(OwnedRoomId),
+    UserId(OwnedUserId),
+    DisplayName(String, OwnedUserId),
 }
 
 impl StyleTreeNode {
-    pub fn to_text(&self, width: usize, style: Style, emoji_shortcodes: bool) -> Text {
-        let mut printer = TextPrinter::new(width, style, true, emoji_shortcodes);
+    pub fn to_text<'a>(
+        &'a self,
+        width: usize,
+        style: Style,
+        settings: &'a ApplicationSettings,
+    ) -> Text<'a> {
+        let mut printer = TextPrinter::new(width, style, true, settings);
         self.print(&mut printer, style);
         printer.finish()
     }
@@ -312,6 +331,12 @@ impl StyleTreeNode {
             StyleTreeNode::Ruler => {},
             StyleTreeNode::Text(_) => {},
             StyleTreeNode::Break => {},
+
+            // TODO: eventually these should turn into internal links:
+            StyleTreeNode::UserId(_) => {},
+            StyleTreeNode::RoomId(_) => {},
+            StyleTreeNode::RoomAlias(_) => {},
+            StyleTreeNode::DisplayName(_, _) => {},
         }
     }
 
@@ -328,11 +353,14 @@ impl StyleTreeNode {
                 printer.push_span_nobreak(span);
             },
             StyleTreeNode::Blockquote(child) => {
-                let mut subp = printer.sub(4);
+                let mut subp = printer.sub(3);
                 child.print(&mut subp, style);
 
                 for mut line in subp.finish() {
-                    line.spans.insert(0, Span::styled("    ", style));
+                    line.spans.insert(0, Span::styled(" ", style));
+                    line.spans
+                        .insert(0, Span::styled(line::THICK_VERTICAL, style.fg(QUOTE_COLOR)));
+                    line.spans.insert(0, Span::styled(" ", style));
                     printer.push_line(line);
                 }
             },
@@ -430,14 +458,14 @@ impl StyleTreeNode {
                 }
             },
             StyleTreeNode::Table(table) => {
-                let text = table.to_text(width, style, printer.emoji_shortcodes());
+                let text = table.to_text(width, style, printer.settings);
                 printer.push_text(text);
             },
             StyleTreeNode::Break => {
                 printer.push_break();
             },
             StyleTreeNode::Text(s) => {
-                printer.push_str(s.as_str(), style);
+                printer.push_str(s.as_ref(), style);
             },
 
             StyleTreeNode::Style(child, patch) => child.print(printer, style.patch(*patch)),
@@ -446,13 +474,30 @@ impl StyleTreeNode {
                     child.print(printer, style);
                 }
             },
+
+            StyleTreeNode::UserId(user_id) => {
+                let style = printer.settings().get_user_style(user_id);
+                printer.push_str(user_id.as_str(), style);
+            },
+            StyleTreeNode::DisplayName(display_name, user_id) => {
+                let style = printer.settings().get_user_style(user_id);
+                printer.push_str(display_name.as_str(), style);
+            },
+            StyleTreeNode::RoomId(room_id) => {
+                let bold = style.add_modifier(StyleModifier::BOLD);
+                printer.push_str(room_id.as_str(), bold);
+            },
+            StyleTreeNode::RoomAlias(alias) => {
+                let bold = style.add_modifier(StyleModifier::BOLD);
+                printer.push_str(alias.as_str(), bold);
+            },
         }
     }
 }
 
 /// A processed HTML document.
 pub struct StyleTree {
-    children: StyleTreeChildren,
+    pub(super) children: StyleTreeChildren,
 }
 
 impl StyleTree {
@@ -466,14 +511,14 @@ impl StyleTree {
         return links;
     }
 
-    pub fn to_text(
-        &self,
+    pub fn to_text<'a>(
+        &'a self,
         width: usize,
         style: Style,
         hide_reply: bool,
-        emoji_shortcodes: bool,
-    ) -> Text<'_> {
-        let mut printer = TextPrinter::new(width, style, hide_reply, emoji_shortcodes);
+        settings: &'a ApplicationSettings,
+    ) -> Text<'a> {
+        let mut printer = TextPrinter::new(width, style, hide_reply, settings);
 
         for child in self.children.iter() {
             child.print(&mut printer, style);
@@ -484,11 +529,11 @@ impl StyleTree {
 }
 
 pub struct TreeGenState {
-    link_num: u8,
+    pub link_num: u8,
 }
 
 impl TreeGenState {
-    fn next_link_char(&mut self) -> Option<char> {
+    pub fn next_link_char(&mut self) -> Option<char> {
         let num = self.link_num;
 
         if num < 62 {
@@ -661,7 +706,7 @@ fn h2t(hdl: &Handle, state: &mut TreeGenState) -> StyleTreeChildren {
 
     let tree = match &node.data {
         NodeData::Document => *c2t(node.children.borrow().as_slice(), state),
-        NodeData::Text { contents } => StyleTreeNode::Text(contents.borrow().to_string()),
+        NodeData::Text { contents } => StyleTreeNode::Text(contents.borrow().to_string().into()),
         NodeData::Element { name, attrs, .. } => {
             match name.local.as_ref() {
                 // Message that this one replies to.
@@ -708,7 +753,7 @@ fn h2t(hdl: &Handle, state: &mut TreeGenState) -> StyleTreeChildren {
 
                     StyleTreeNode::Style(c, s)
                 },
-                "del" | "strike" => {
+                "del" | "s" | "strike" => {
                     let c = c2t(&node.children.borrow(), state);
                     let s = Style::default().add_modifier(StyleModifier::CROSSED_OUT);
 
@@ -775,7 +820,8 @@ fn h2t(hdl: &Handle, state: &mut TreeGenState) -> StyleTreeChildren {
                     *c2t(&node.children.borrow(), state)
                 },
 
-                _ => return vec![],
+                // Treat unknown tags as plain text.
+                _ => *c2t(&node.children.borrow(), state),
             }
         },
 
@@ -811,17 +857,19 @@ pub fn parse_matrix_html(s: &str) -> StyleTree {
 #[cfg(test)]
 pub mod tests {
     use super::*;
+    use crate::tests::mock_settings;
     use crate::util::space_span;
     use pretty_assertions::assert_eq;
     use unicode_width::UnicodeWidthStr;
 
     #[test]
     fn test_header() {
+        let settings = mock_settings();
         let bold = Style::default().add_modifier(StyleModifier::BOLD);
 
         let s = "<h1>Header 1</h1>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("#", bold),
             Span::styled(" ", bold),
@@ -833,7 +881,7 @@ pub mod tests {
 
         let s = "<h2>Header 2</h2>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("#", bold),
             Span::styled("#", bold),
@@ -846,7 +894,7 @@ pub mod tests {
 
         let s = "<h3>Header 3</h3>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("#", bold),
             Span::styled("#", bold),
@@ -860,7 +908,7 @@ pub mod tests {
 
         let s = "<h4>Header 4</h4>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("#", bold),
             Span::styled("#", bold),
@@ -875,7 +923,7 @@ pub mod tests {
 
         let s = "<h5>Header 5</h5>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("#", bold),
             Span::styled("#", bold),
@@ -891,7 +939,7 @@ pub mod tests {
 
         let s = "<h6>Header 6</h6>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("#", bold),
             Span::styled("#", bold),
@@ -909,6 +957,7 @@ pub mod tests {
 
     #[test]
     fn test_style() {
+        let settings = mock_settings();
         let def = Style::default();
         let bold = def.add_modifier(StyleModifier::BOLD);
         let italic = def.add_modifier(StyleModifier::ITALIC);
@@ -918,7 +967,7 @@ pub mod tests {
 
         let s = "<b>Bold!</b>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Bold", bold),
             Span::styled("!", bold),
@@ -927,7 +976,7 @@ pub mod tests {
 
         let s = "<strong>Bold!</strong>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Bold", bold),
             Span::styled("!", bold),
@@ -936,7 +985,7 @@ pub mod tests {
 
         let s = "<i>Italic!</i>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Italic", italic),
             Span::styled("!", italic),
@@ -945,7 +994,7 @@ pub mod tests {
 
         let s = "<em>Italic!</em>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Italic", italic),
             Span::styled("!", italic),
@@ -954,7 +1003,7 @@ pub mod tests {
 
         let s = "<del>Strikethrough!</del>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Strikethrough", strike),
             Span::styled("!", strike),
@@ -963,7 +1012,7 @@ pub mod tests {
 
         let s = "<strike>Strikethrough!</strike>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Strikethrough", strike),
             Span::styled("!", strike),
@@ -972,7 +1021,7 @@ pub mod tests {
 
         let s = "<u>Underline!</u>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Underline", underl),
             Span::styled("!", underl),
@@ -981,7 +1030,7 @@ pub mod tests {
 
         let s = "<font color=\"#ff0000\">Red!</u>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Red", red),
             Span::styled("!", red),
@@ -990,7 +1039,7 @@ pub mod tests {
 
         let s = "<font color=\"red\">Red!</u>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(20, Style::default(), false, false);
+        let text = tree.to_text(20, Style::default(), false, &settings);
         assert_eq!(text.lines, vec![Line::from(vec![
             Span::styled("Red", red),
             Span::styled("!", red),
@@ -1000,9 +1049,10 @@ pub mod tests {
 
     #[test]
     fn test_paragraph() {
+        let settings = mock_settings();
         let s = "<p>Hello world!</p><p>Content</p><p>Goodbye world!</p>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(10, Style::default(), false, false);
+        let text = tree.to_text(10, Style::default(), false, &settings);
         assert_eq!(text.lines.len(), 7);
         assert_eq!(
             text.lines[0],
@@ -1027,25 +1077,42 @@ pub mod tests {
 
     #[test]
     fn test_blockquote() {
+        let settings = mock_settings();
         let s = "<blockquote>Hello world!</blockquote>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(10, Style::default(), false, false);
+        let text = tree.to_text(10, Style::default(), false, &settings);
+        let style = Style::new().fg(QUOTE_COLOR);
         assert_eq!(text.lines.len(), 2);
         assert_eq!(
             text.lines[0],
-            Line::from(vec![Span::raw("    "), Span::raw("Hello"), Span::raw(" ")])
+            Line::from(vec![
+                Span::raw(" "),
+                Span::styled(line::THICK_VERTICAL, style),
+                Span::raw(" "),
+                Span::raw("Hello"),
+                Span::raw(" "),
+                Span::raw(" "),
+            ])
         );
         assert_eq!(
             text.lines[1],
-            Line::from(vec![Span::raw("    "), Span::raw("world"), Span::raw("!")])
+            Line::from(vec![
+                Span::raw(" "),
+                Span::styled(line::THICK_VERTICAL, style),
+                Span::raw(" "),
+                Span::raw("world"),
+                Span::raw("!"),
+                Span::raw(" "),
+            ])
         );
     }
 
     #[test]
     fn test_list_unordered() {
+        let settings = mock_settings();
         let s = "<ul><li>List Item 1</li><li>List Item 2</li><li>List Item 3</li></ul>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(8, Style::default(), false, false);
+        let text = tree.to_text(8, Style::default(), false, &settings);
         assert_eq!(text.lines.len(), 6);
         assert_eq!(
             text.lines[0],
@@ -1105,9 +1172,10 @@ pub mod tests {
 
     #[test]
     fn test_list_ordered() {
+        let settings = mock_settings();
         let s = "<ol><li>List Item 1</li><li>List Item 2</li><li>List Item 3</li></ol>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(9, Style::default(), false, false);
+        let text = tree.to_text(9, Style::default(), false, &settings);
         assert_eq!(text.lines.len(), 6);
         assert_eq!(
             text.lines[0],
@@ -1167,6 +1235,7 @@ pub mod tests {
 
     #[test]
     fn test_table() {
+        let settings = mock_settings();
         let s = "<table>\
                  <thead>\
                  <tr><th>Column 1</th><th>Column 2</th><th>Column 3</th></tr>
@@ -1177,7 +1246,7 @@ pub mod tests {
                  <tr><td>a</td><td>b</td><td>c</td></tr>\
                  </tbody></table>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(15, Style::default(), false, false);
+        let text = tree.to_text(15, Style::default(), false, &settings);
         let bold = Style::default().add_modifier(StyleModifier::BOLD);
         assert_eq!(text.lines.len(), 11);
 
@@ -1267,10 +1336,11 @@ pub mod tests {
 
     #[test]
     fn test_matrix_reply() {
+        let settings = mock_settings();
         let s = "<mx-reply>This was replied to</mx-reply>This is the reply";
 
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(10, Style::default(), false, false);
+        let text = tree.to_text(10, Style::default(), false, &settings);
         assert_eq!(text.lines.len(), 4);
         assert_eq!(
             text.lines[0],
@@ -1307,7 +1377,7 @@ pub mod tests {
         );
 
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(10, Style::default(), true, false);
+        let text = tree.to_text(10, Style::default(), true, &settings);
         assert_eq!(text.lines.len(), 2);
         assert_eq!(
             text.lines[0],
@@ -1332,9 +1402,10 @@ pub mod tests {
 
     #[test]
     fn test_self_closing() {
+        let settings = mock_settings();
         let s = "Hello<br>World<br>Goodbye";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(7, Style::default(), true, false);
+        let text = tree.to_text(7, Style::default(), true, &settings);
         assert_eq!(text.lines.len(), 3);
         assert_eq!(text.lines[0], Line::from(vec![Span::raw("Hello"), Span::raw("  "),]));
         assert_eq!(text.lines[1], Line::from(vec![Span::raw("World"), Span::raw("  "),]));
@@ -1343,9 +1414,10 @@ pub mod tests {
 
     #[test]
     fn test_embedded_newline() {
+        let settings = mock_settings();
         let s = "<p>Hello\nWorld</p>";
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(15, Style::default(), true, false);
+        let text = tree.to_text(15, Style::default(), true, &settings);
         assert_eq!(text.lines.len(), 1);
         assert_eq!(
             text.lines[0],
@@ -1360,16 +1432,18 @@ pub mod tests {
 
     #[test]
     fn test_pre_tag() {
+        let settings = mock_settings();
         let s = concat!(
             "<pre><code class=\"language-rust\">",
             "fn hello() -&gt; usize {\n",
+            " \t// weired\n",
             "    return 5;\n",
             "}\n",
             "</code></pre>\n"
         );
         let tree = parse_matrix_html(s);
-        let text = tree.to_text(25, Style::default(), true, false);
-        assert_eq!(text.lines.len(), 5);
+        let text = tree.to_text(25, Style::default(), true, &settings);
+        assert_eq!(text.lines.len(), 6);
         assert_eq!(
             text.lines[0],
             Line::from(vec![
@@ -1400,6 +1474,20 @@ pub mod tests {
         );
         assert_eq!(
             text.lines[2],
+            Line::from(vec![
+                Span::raw(line::VERTICAL),
+                Span::raw(" "),
+                Span::raw("   "),
+                Span::raw("/"),
+                Span::raw("/"),
+                Span::raw(" "),
+                Span::raw("weired"),
+                Span::raw("          "),
+                Span::raw(line::VERTICAL)
+            ])
+        );
+        assert_eq!(
+            text.lines[3],
             Line::from(vec![
                 Span::raw(line::VERTICAL),
                 Span::raw("    "),
@@ -1412,7 +1500,7 @@ pub mod tests {
             ])
         );
         assert_eq!(
-            text.lines[3],
+            text.lines[4],
             Line::from(vec![
                 Span::raw(line::VERTICAL),
                 Span::raw("}"),
@@ -1421,7 +1509,7 @@ pub mod tests {
             ])
         );
         assert_eq!(
-            text.lines[4],
+            text.lines[5],
             Line::from(vec![
                 Span::raw(line::BOTTOM_LEFT),
                 Span::raw(line::HORIZONTAL.repeat(23)),
@@ -1432,6 +1520,11 @@ pub mod tests {
 
     #[test]
     fn test_emoji_shortcodes() {
+        let mut enabled = mock_settings();
+        enabled.tunables.message_shortcode_display = true;
+        let mut disabled = mock_settings();
+        disabled.tunables.message_shortcode_display = false;
+
         for shortcode in ["exploding_head", "polar_bear", "canada"] {
             let emoji = emojis::get_by_shortcode(shortcode).unwrap().as_str();
             let emoji_width = UnicodeWidthStr::width(emoji);
@@ -1440,13 +1533,13 @@ pub mod tests {
             let s = format!("<p>{emoji}</p>");
             let tree = parse_matrix_html(s.as_str());
             // Test with emojis_shortcodes set to false
-            let text = tree.to_text(20, Style::default(), false, false);
+            let text = tree.to_text(20, Style::default(), false, &disabled);
             assert_eq!(text.lines, vec![Line::from(vec![
                 Span::raw(emoji),
                 space_span(20 - emoji_width, Style::default()),
             ]),]);
             // Test with emojis_shortcodes set to true
-            let text = tree.to_text(20, Style::default(), false, true);
+            let text = tree.to_text(20, Style::default(), false, &enabled);
             assert_eq!(text.lines, vec![Line::from(vec![
                 Span::raw(replacement.as_str()),
                 space_span(20 - replacement_width, Style::default()),
