@@ -2,15 +2,16 @@
 use std::borrow::Cow;
 use std::cmp::{Ord, Ordering, PartialOrd};
 use std::collections::hash_map::DefaultHasher;
-use std::collections::hash_set;
 use std::collections::BTreeMap;
 use std::convert::{TryFrom, TryInto};
 use std::fmt::{self, Display};
 use std::hash::{Hash, Hasher};
 use std::ops::{Deref, DerefMut};
 
-use chrono::{DateTime, Local as LocalTz, NaiveDateTime, TimeZone};
+use chrono::{DateTime, Local as LocalTz};
 use humansize::{format_size, DECIMAL};
+use matrix_sdk::ruma::events::receipt::ReceiptThread;
+use matrix_sdk::ruma::room_version_rules::RedactionRules;
 use serde_json::json;
 use unicode_width::UnicodeWidthStr;
 
@@ -35,6 +36,7 @@ use matrix_sdk::ruma::{
             },
             redaction::SyncRoomRedactionEvent,
         },
+        AnySyncStateEvent,
         RedactContent,
         RedactedUnsigned,
     },
@@ -42,7 +44,6 @@ use matrix_sdk::ruma::{
     MilliSecondsSinceUnixEpoch,
     OwnedEventId,
     OwnedUserId,
-    RoomVersionId,
     UInt,
 };
 
@@ -67,13 +68,17 @@ use crate::{
 mod compose;
 mod html;
 mod printer;
+mod state;
 
 pub use self::compose::text_to_message;
+use self::state::{body_cow_state, html_state};
+pub use html::TreeGenState;
+
+type ProtocolPreview<'a> = (&'a Protocol, u16, u16);
 
 pub type MessageKey = (MessageTimeStamp, OwnedEventId);
 
-#[derive(Default)]
-pub struct Messages(BTreeMap<MessageKey, Message>);
+pub struct Messages(BTreeMap<MessageKey, Message>, pub ReceiptThread);
 
 impl Deref for Messages {
     type Target = BTreeMap<MessageKey, Message>;
@@ -90,6 +95,18 @@ impl DerefMut for Messages {
 }
 
 impl Messages {
+    pub fn new(thread: ReceiptThread) -> Self {
+        Self(Default::default(), thread)
+    }
+
+    pub fn main() -> Self {
+        Self::new(ReceiptThread::Main)
+    }
+
+    pub fn thread(root: OwnedEventId) -> Self {
+        Self::new(ReceiptThread::Thread(root))
+    }
+
     pub fn insert_message(&mut self, key: MessageKey, msg: impl Into<Message>) {
         let event_id = key.1.clone();
         let msg = msg.into();
@@ -155,12 +172,15 @@ fn placeholder_frame(
     image_preview_size: &ImagePreviewSize,
 ) -> Option<String> {
     let ImagePreviewSize { width, height } = image_preview_size;
-    if outer_width < *width || (*width < 2 || *height < 2) {
+    let width = usize::min(*width, outer_width);
+    if width < 2 || *height < 2 {
         return None;
     }
     let mut placeholder = "\u{230c}".to_string();
     placeholder.push_str(&" ".repeat(width - 2));
-    placeholder.push_str("\u{230d}\n");
+    placeholder.push('\u{230d}');
+    placeholder.push_str(&"\n".repeat((height - 1) / 2));
+
     if *height > 2 {
         if let Some(text) = text {
             if text.width() <= width - 2 {
@@ -170,7 +190,7 @@ fn placeholder_frame(
         }
     }
 
-    placeholder.push_str(&"\n".repeat(height - 2));
+    placeholder.push_str(&"\n".repeat(height / 2));
     placeholder.push('\u{230e}');
     placeholder.push_str(&" ".repeat(width - 2));
     placeholder.push_str("\u{230f}\n");
@@ -180,9 +200,8 @@ fn placeholder_frame(
 #[inline]
 fn millis_to_datetime(ms: UInt) -> DateTime<LocalTz> {
     let time = i64::from(ms) / 1000;
-    let time = NaiveDateTime::from_timestamp_opt(time, 0).unwrap_or_default();
-
-    LocalTz.from_utc_datetime(&time)
+    let time = DateTime::from_timestamp(time, 0).unwrap_or_default();
+    time.into()
 }
 
 #[derive(thiserror::Error, Debug)]
@@ -215,13 +234,13 @@ impl MessageTimeStamp {
         dt1.date_naive() == dt2.date_naive()
     }
 
-    fn show_date(&self) -> Option<Span> {
+    fn show_date(&self) -> Option<Span<'_>> {
         let time = self.as_datetime().format("%A, %B %d %Y").to_string();
 
         Span::styled(time, BOLD_STYLE).into()
     }
 
-    fn show_time(&self) -> Option<Span> {
+    fn show_time(&self) -> Option<Span<'_>> {
         match self {
             MessageTimeStamp::OriginServer(ms) => {
                 let time = millis_to_datetime(*ms).format("%T");
@@ -426,6 +445,7 @@ pub enum MessageEvent {
     EncryptedRedacted(Box<RedactedRoomEncryptedEvent>),
     Original(Box<OriginalRoomMessageEvent>),
     Redacted(Box<RedactedRoomMessageEvent>),
+    State(Box<AnySyncStateEvent>),
     Local(OwnedEventId, Box<RoomMessageEventContent>),
 }
 
@@ -436,6 +456,7 @@ impl MessageEvent {
             MessageEvent::EncryptedRedacted(ev) => ev.event_id.as_ref(),
             MessageEvent::Original(ev) => ev.event_id.as_ref(),
             MessageEvent::Redacted(ev) => ev.event_id.as_ref(),
+            MessageEvent::State(ev) => ev.event_id(),
             MessageEvent::Local(event_id, _) => event_id.as_ref(),
         }
     }
@@ -446,6 +467,7 @@ impl MessageEvent {
             MessageEvent::Original(ev) => Some(&ev.content),
             MessageEvent::EncryptedRedacted(_) => None,
             MessageEvent::Redacted(_) => None,
+            MessageEvent::State(_) => None,
             MessageEvent::Local(_, content) => Some(content),
         }
     }
@@ -463,6 +485,7 @@ impl MessageEvent {
             MessageEvent::Original(ev) => body_cow_content(&ev.content),
             MessageEvent::EncryptedRedacted(ev) => body_cow_reason(&ev.unsigned),
             MessageEvent::Redacted(ev) => body_cow_reason(&ev.unsigned),
+            MessageEvent::State(ev) => body_cow_state(ev),
             MessageEvent::Local(_, content) => body_cow_content(content),
         }
     }
@@ -473,6 +496,7 @@ impl MessageEvent {
             MessageEvent::EncryptedRedacted(_) => return None,
             MessageEvent::Original(ev) => &ev.content,
             MessageEvent::Redacted(_) => return None,
+            MessageEvent::State(ev) => return Some(html_state(ev)),
             MessageEvent::Local(_, content) => content,
         };
 
@@ -487,15 +511,16 @@ impl MessageEvent {
         }
     }
 
-    fn redact(&mut self, redaction: SyncRoomRedactionEvent, version: &RoomVersionId) {
+    fn redact(&mut self, redaction: SyncRoomRedactionEvent, rules: &RedactionRules) {
         match self {
             MessageEvent::EncryptedOriginal(_) => return,
             MessageEvent::EncryptedRedacted(_) => return,
             MessageEvent::Redacted(_) => return,
+            MessageEvent::State(_) => return,
             MessageEvent::Local(_, _) => return,
             MessageEvent::Original(ev) => {
                 let redacted = RedactedRoomMessageEvent {
-                    content: ev.content.clone().redact(version),
+                    content: ev.content.clone().redact(rules),
                     event_id: ev.event_id.clone(),
                     sender: ev.sender.clone(),
                     origin_server_ts: ev.origin_server_ts,
@@ -549,27 +574,18 @@ fn body_cow_content(content: &RoomMessageEventContent) -> Cow<'_, str> {
         MessageType::Video(content) => {
             display_file_to_text!(Video, content);
         },
-        _ => {
-            match content.msgtype() {
-                // Just show the body text for the special Element messages.
-                "nic.custom.confetti" |
-                "nic.custom.fireworks" |
-                "io.element.effect.hearts" |
-                "io.element.effect.rainfall" |
-                "io.element.effect.snowfall" |
-                "io.element.effects.space_invaders" => content.body(),
-                other => {
-                    return Cow::Owned(format!("[Unknown message type: {other:?}]"));
-                },
-            }
-        },
+        _ => content.body(),
     };
 
     Cow::Borrowed(s)
 }
 
 fn body_cow_reason(unsigned: &RedactedUnsigned) -> Cow<'_, str> {
-    let reason = unsigned.redacted_because.content.reason.as_ref();
+    let reason = unsigned
+        .redacted_because
+        .deserialize()
+        .ok()
+        .and_then(|ev| ev.content.reason);
 
     if let Some(r) = reason {
         Cow::Owned(format!("[Redacted: {r:?}]"))
@@ -623,8 +639,8 @@ struct MessageFormatter<'a> {
     /// The date the message was sent.
     date: Option<Span<'a>>,
 
-    /// Iterator over the users who have read up to this message.
-    read: Option<hash_set::Iter<'a, OwnedUserId>>,
+    /// The users who have read up to this message.
+    read: Vec<OwnedUserId>,
 }
 
 impl<'a> MessageFormatter<'a> {
@@ -657,13 +673,11 @@ impl<'a> MessageFormatter<'a> {
                 line.push(time);
 
                 // Show read receipts.
-                let user_char =
-                    |user: &'a OwnedUserId| -> Span<'a> { settings.get_user_char_span(user) };
-                let mut read = self.read.iter_mut().flatten();
+                let user_char = |user: OwnedUserId| -> Span { settings.get_user_char_span(&user) };
 
-                let a = read.next().map(user_char).unwrap_or_else(|| Span::raw(" "));
-                let b = read.next().map(user_char).unwrap_or_else(|| Span::raw(" "));
-                let c = read.next().map(user_char).unwrap_or_else(|| Span::raw(" "));
+                let a = self.read.pop().map(user_char).unwrap_or_else(|| Span::raw(" "));
+                let b = self.read.pop().map(user_char).unwrap_or_else(|| Span::raw(" "));
+                let c = self.read.pop().map(user_char).unwrap_or_else(|| Span::raw(" "));
 
                 line.push(Span::raw(" "));
                 line.push(c);
@@ -716,39 +730,55 @@ impl<'a> MessageFormatter<'a> {
         style: Style,
         text: &mut Text<'a>,
         info: &'a RoomInfo,
-    ) {
+        settings: &'a ApplicationSettings,
+    ) -> Option<ProtocolPreview<'a>> {
+        let reply_style = if settings.tunables.message_user_color {
+            style.patch(settings.get_user_color(&msg.sender))
+        } else {
+            style
+        };
+
         let width = self.width();
         let w = width.saturating_sub(2);
-        let shortcodes = self.settings.tunables.message_shortcode_display;
-        let (mut replied, _) = msg.show_msg(w, style, true, shortcodes);
+        let (mut replied, proto) = msg.show_msg(w, reply_style, true, settings);
         let mut sender = msg.sender_span(info, self.settings);
         let sender_width = UnicodeWidthStr::width(sender.content.as_ref());
         let trailing = w.saturating_sub(sender_width + 1);
 
-        sender.style = sender.style.patch(style);
+        sender.style = sender.style.patch(reply_style);
 
         self.push_spans(
             Line::from(vec![
                 Span::styled(" ", style),
                 Span::styled(THICK_VERTICAL, style),
                 sender,
-                Span::styled(":", style),
-                space_span(trailing, style),
+                Span::styled(":", reply_style),
+                space_span(trailing, reply_style),
             ]),
             style,
             text,
         );
 
+        // Determine the image offset of the reply header, taking into account the formatting
+        let proto = proto.map(|p| {
+            let y_off = text.lines.len() as u16;
+            // Adjust x_off by 2 to account for the vertical line and indent
+            let x_off = self.cols.user_gutter_width(settings) + 2;
+            (p, x_off, y_off)
+        });
+
         for line in replied.lines.iter_mut() {
             line.spans.insert(0, Span::styled(THICK_VERTICAL, style));
             line.spans.insert(0, Span::styled(" ", style));
         }
 
-        self.push_text(replied, style, text);
+        self.push_text(replied, reply_style, text);
+
+        proto
     }
 
     fn push_reactions(&mut self, counts: Vec<(&'a str, usize)>, style: Style, text: &mut Text<'a>) {
-        let mut emojis = printer::TextPrinter::new(self.width(), style, false, false);
+        let mut emojis = printer::TextPrinter::new(self.width(), style, false, self.settings);
         let mut reactions = 0;
 
         for (key, count) in counts {
@@ -797,7 +827,7 @@ impl<'a> MessageFormatter<'a> {
         let plural = len != 1;
         let style = Style::default();
         let mut threaded =
-            printer::TextPrinter::new(self.width(), style, false, false).literal(true);
+            printer::TextPrinter::new(self.width(), style, false, self.settings).literal(true);
         let len = Span::styled(len.to_string(), style.add_modifier(StyleModifier::BOLD));
         threaded.push_str(" \u{2937} ", style);
         threaded.push_span_nobreak(len);
@@ -814,7 +844,7 @@ impl<'a> MessageFormatter<'a> {
 pub enum ImageStatus {
     None,
     Downloading(ImagePreviewSize),
-    Loaded(Box<dyn Protocol>),
+    Loaded(Protocol),
     Error(String),
 }
 
@@ -849,6 +879,7 @@ impl Message {
             MessageEvent::Local(_, content) => content,
             MessageEvent::Original(ev) => &ev.content,
             MessageEvent::Redacted(_) => return None,
+            MessageEvent::State(_) => return None,
         };
 
         match &content.relates_to {
@@ -869,6 +900,7 @@ impl Message {
             MessageEvent::Local(_, content) => content,
             MessageEvent::Original(ev) => &ev.content,
             MessageEvent::Redacted(_) => return None,
+            MessageEvent::State(_) => return None,
         };
 
         match &content.relates_to {
@@ -922,7 +954,13 @@ impl Message {
             let fill = width - user_gutter - TIME_GUTTER - READ_GUTTER;
             let user = self.show_sender(prev, true, info, settings);
             let time = self.timestamp.show_time();
-            let read = info.event_receipts.get(self.event.event_id()).map(|read| read.iter());
+            let read = info
+                .event_receipts
+                .values()
+                .filter_map(|receipts| receipts.get(self.event.event_id()))
+                .flat_map(|read| read.iter())
+                .map(|user_id| user_id.to_owned())
+                .collect();
 
             MessageFormatter { settings, cols, orig, fill, user, date, time, read }
         } else if user_gutter + TIME_GUTTER + MIN_MSG_LEN <= width {
@@ -930,7 +968,7 @@ impl Message {
             let fill = width - user_gutter - TIME_GUTTER;
             let user = self.show_sender(prev, true, info, settings);
             let time = self.timestamp.show_time();
-            let read = None;
+            let read = Vec::new();
 
             MessageFormatter { settings, cols, orig, fill, user, date, time, read }
         } else if user_gutter + MIN_MSG_LEN <= width {
@@ -938,7 +976,7 @@ impl Message {
             let fill = width - user_gutter;
             let user = self.show_sender(prev, true, info, settings);
             let time = None;
-            let read = None;
+            let read = Vec::new();
 
             MessageFormatter { settings, cols, orig, fill, user, date, time, read }
         } else {
@@ -946,7 +984,7 @@ impl Message {
             let fill = width.saturating_sub(2);
             let user = self.show_sender(prev, false, info, settings);
             let time = None;
-            let read = None;
+            let read = Vec::new();
 
             MessageFormatter { settings, cols, orig, fill, user, date, time, read }
         }
@@ -962,7 +1000,7 @@ impl Message {
         vwctx: &ViewportContext<MessageCursor>,
         info: &'a RoomInfo,
         settings: &'a ApplicationSettings,
-    ) -> (Text<'a>, Option<(&dyn Protocol, u16, u16)>) {
+    ) -> (Text<'a>, [Option<ProtocolPreview<'a>>; 2]) {
         let width = vwctx.get_width();
 
         let style = self.get_render_style(selected, settings);
@@ -975,24 +1013,20 @@ impl Message {
             .reply_to()
             .or_else(|| self.thread_root())
             .and_then(|e| info.get_event(&e));
-
-        if let Some(r) = &reply {
-            fmt.push_in_reply(r, style, &mut text, info);
-        }
+        let proto_reply = reply.as_ref().and_then(|r| {
+            // Format the reply header, push it into the `Text` buffer, and get any image.
+            fmt.push_in_reply(r, style, &mut text, info, settings)
+        });
 
         // Now show the message contents, and the inlined reply if we couldn't find it above.
-        let (msg, proto) = self.show_msg(
-            width,
-            style,
-            reply.is_some(),
-            settings.tunables.message_shortcode_display,
-        );
+        let (msg, proto) = self.show_msg(width, style, reply.is_some(), settings);
 
         // Given our text so far, determine the image offset.
-        let proto = proto.map(|p| {
+        let proto_main = proto.map(|p| {
             let y_off = text.lines.len() as u16;
             let x_off = fmt.cols.user_gutter_width(settings);
-            // Adjust y_off by 1 if a date was printed before the message to account for the extra line.
+            // Adjust y_off by 1 if a date was printed before the message to account for
+            // the extra line we're going to print.
             let y_off = if fmt.date.is_some() { y_off + 1 } else { y_off };
             (p, x_off, y_off)
         });
@@ -1013,7 +1047,7 @@ impl Message {
             fmt.push_thread_reply_count(thread.len(), &mut text);
         }
 
-        (text, proto)
+        (text, [proto_main, proto_reply])
     }
 
     pub fn show<'a>(
@@ -1027,18 +1061,18 @@ impl Message {
         self.show_with_preview(prev, selected, vwctx, info, settings).0
     }
 
-    fn show_msg(
-        &self,
+    fn show_msg<'a>(
+        &'a self,
         width: usize,
         style: Style,
         hide_reply: bool,
-        emoji_shortcodes: bool,
-    ) -> (Text, Option<&dyn Protocol>) {
+        settings: &'a ApplicationSettings,
+    ) -> (Text<'a>, Option<&'a Protocol>) {
         if let Some(html) = &self.html {
-            (html.to_text(width, style, hide_reply, emoji_shortcodes), None)
+            (html.to_text(width, style, hide_reply, settings), None)
         } else {
             let mut msg = self.event.body();
-            if emoji_shortcodes {
+            if settings.tunables.message_shortcode_display {
                 msg = Cow::Owned(replace_emojis_in_str(msg.as_ref()));
             }
 
@@ -1053,8 +1087,8 @@ impl Message {
                     placeholder_frame(Some("Downloading..."), width, image_preview_size)
                 },
                 ImageStatus::Loaded(backend) => {
-                    proto = Some(backend.as_ref());
-                    placeholder_frame(None, width, &backend.rect().into())
+                    proto = Some(backend);
+                    placeholder_frame(Some("No Space..."), width, &backend.area().into())
                 },
                 ImageStatus::Error(err) => Some(format!("[Image error: {err}]\n")),
             };
@@ -1097,17 +1131,19 @@ impl Message {
         let padding = user_gutter - 2 - width;
 
         let sender = if align_right {
-            space(padding) + &truncated + "  "
+            format!("{}{}  ", space(padding), truncated)
         } else {
-            truncated.into_owned() + &space(padding) + "  "
+            format!("{}{}  ", truncated, space(padding))
         };
 
         Span::styled(sender, style).into()
     }
 
-    pub fn redact(&mut self, redaction: SyncRoomRedactionEvent, version: &RoomVersionId) {
-        self.event.redact(redaction, version);
+    pub fn redact(&mut self, redaction: SyncRoomRedactionEvent, rules: &RedactionRules) {
+        self.event.redact(redaction, rules);
         self.html = None;
+        self.downloaded = false;
+        self.image_preview = ImageStatus::None;
     }
 }
 
@@ -1153,6 +1189,16 @@ impl From<RoomMessageEvent> for Message {
     }
 }
 
+impl From<AnySyncStateEvent> for Message {
+    fn from(event: AnySyncStateEvent) -> Self {
+        let timestamp = event.origin_server_ts().into();
+        let user_id = event.sender().to_owned();
+        let event = MessageEvent::State(event.into());
+
+        Message::new(event, user_id, timestamp)
+    }
+}
+
 impl Display for Message {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         write!(f, "{}", self.event.body())
@@ -1251,7 +1297,7 @@ pub mod tests {
         assert_eq!(k6, &MSG1_KEY.clone());
 
         // MessageCursor::latest() fails to convert for a room w/o messages.
-        let messages_empty = Messages::default();
+        let messages_empty = Messages::new(ReceiptThread::Main);
         assert_eq!(mc6.to_key(&messages_empty), None);
     }
 
@@ -1300,7 +1346,17 @@ pub mod tests {
             )
         );
 
-        assert_eq!(placeholder_frame(None, 2, &ImagePreviewSize { width: 4, height: 4 }), None);
+        assert_eq!(
+            placeholder_frame(None, 2, &ImagePreviewSize { width: 4, height: 4 }),
+            pretty_frame_test(
+                r#"
+⌌⌍
+
+
+⌎⌏
+"#
+            )
+        );
         assert_eq!(placeholder_frame(None, 4, &ImagePreviewSize { width: 1, height: 4 }), None);
 
         assert_eq!(placeholder_frame(None, 4, &ImagePreviewSize { width: 4, height: 1 }), None);
@@ -1313,6 +1369,33 @@ pub mod tests {
  OK
 
 ⌎  ⌏
+"#
+            )
+        );
+        assert_eq!(
+            placeholder_frame(Some("OK"), 6, &ImagePreviewSize { width: 6, height: 6 }),
+            pretty_frame_test(
+                r#"
+⌌    ⌍
+
+ OK
+
+
+⌎    ⌏
+"#
+            )
+        );
+        assert_eq!(
+            placeholder_frame(Some("OK"), 6, &ImagePreviewSize { width: 6, height: 7 }),
+            pretty_frame_test(
+                r#"
+⌌    ⌍
+
+
+ OK
+
+
+⌎    ⌏
 "#
             )
         );
