@@ -5,7 +5,7 @@
 //!
 //! Additionally, some of the iamb commands delegate behaviour to the current UI element. For
 //! example, [sending messages][crate::base::SendAction] delegate to the [room window][RoomState],
-//! where we have the message bar and room ID easily accesible and resetable.
+//! where we have the message bar and room ID easily accessible and resettable.
 use std::cmp::{Ord, Ordering, PartialOrd};
 use std::fmt::{self, Display};
 use std::ops::Deref;
@@ -23,6 +23,7 @@ use matrix_sdk::{
         RoomAliasId,
         RoomId,
     },
+    RoomState as MatrixRoomState,
 };
 
 use ratatui::{
@@ -65,7 +66,6 @@ use crate::base::{
     IambInfo,
     IambResult,
     MessageAction,
-    Need,
     ProgramAction,
     ProgramContext,
     ProgramStore,
@@ -75,11 +75,13 @@ use crate::base::{
     SortFieldRoom,
     SortFieldUser,
     SortOrder,
+    SpaceAction,
     UnreadInfo,
 };
 
 use self::{room::RoomState, welcome::WelcomeState};
 use crate::message::MessageTimeStamp;
+use feruca::Collator;
 
 pub mod room;
 pub mod welcome;
@@ -94,12 +96,12 @@ fn bold_style() -> Style {
 }
 
 #[inline]
-fn bold_span(s: &str) -> Span {
+fn bold_span(s: &str) -> Span<'_> {
     Span::styled(s, bold_style())
 }
 
 #[inline]
-fn bold_spans(s: &str) -> Line {
+fn bold_spans(s: &str) -> Line<'_> {
     bold_span(s).into()
 }
 
@@ -113,12 +115,12 @@ fn selected_style(selected: bool) -> Style {
 }
 
 #[inline]
-fn selected_span(s: &str, selected: bool) -> Span {
+fn selected_span(s: &str, selected: bool) -> Span<'_> {
     Span::styled(s, selected_style(selected))
 }
 
 #[inline]
-fn selected_text(s: &str, selected: bool) -> Text {
+fn selected_text(s: &str, selected: bool) -> Text<'_> {
     Text::from(selected_span(s, selected))
 }
 
@@ -168,7 +170,12 @@ fn user_cmp(a: &MemberItem, b: &MemberItem, field: &SortFieldUser) -> Ordering {
     }
 }
 
-fn room_cmp<T: RoomLikeItem>(a: &T, b: &T, field: &SortFieldRoom) -> Ordering {
+fn room_cmp<T: RoomLikeItem>(
+    a: &T,
+    b: &T,
+    field: &SortFieldRoom,
+    collator: &mut Collator,
+) -> Ordering {
     match field {
         SortFieldRoom::Favorite => {
             let fava = a.has_tag(TagName::Favorite);
@@ -184,7 +191,7 @@ fn room_cmp<T: RoomLikeItem>(a: &T, b: &T, field: &SortFieldRoom) -> Ordering {
             // If a has LowPriority and b doesn't, it should sort later in room list.
             lowa.cmp(&lowb)
         },
-        SortFieldRoom::Name => a.name().cmp(b.name()),
+        SortFieldRoom::Name => collator.collate(a.name(), b.name()),
         SortFieldRoom::Alias => some_cmp(a.alias(), b.alias(), Ord::cmp),
         SortFieldRoom::RoomId => a.room_id().cmp(b.room_id()),
         SortFieldRoom::Unread => {
@@ -195,6 +202,10 @@ fn room_cmp<T: RoomLikeItem>(a: &T, b: &T, field: &SortFieldRoom) -> Ordering {
             // sort larger timestamps towards the top.
             some_cmp(a.recent_ts(), b.recent_ts(), |a, b| b.cmp(a))
         },
+        SortFieldRoom::Invite => {
+            // sort invites before other rooms.
+            b.is_invite().cmp(&a.is_invite())
+        },
     }
 }
 
@@ -203,9 +214,10 @@ fn room_fields_cmp<T: RoomLikeItem>(
     a: &T,
     b: &T,
     fields: &[SortColumn<SortFieldRoom>],
+    collator: &mut Collator,
 ) -> Ordering {
     for SortColumn(field, order) in fields {
-        match (room_cmp(a, b, field), order) {
+        match (room_cmp(a, b, field, collator), order) {
             (Ordering::Equal, _) => continue,
             (o, SortOrder::Ascending) => return o,
             (o, SortOrder::Descending) => return o.reverse(),
@@ -213,7 +225,7 @@ fn room_fields_cmp<T: RoomLikeItem>(
     }
 
     // Break ties on ascending room id.
-    room_cmp(a, b, &SortFieldRoom::RoomId)
+    room_cmp(a, b, &SortFieldRoom::RoomId, collator)
 }
 
 fn user_fields_cmp(
@@ -273,6 +285,7 @@ trait RoomLikeItem {
     fn recent_ts(&self) -> Option<&MessageTimeStamp>;
     fn alias(&self) -> Option<&RoomAliasId>;
     fn name(&self) -> &str;
+    fn is_invite(&self) -> bool;
 }
 
 #[inline]
@@ -354,6 +367,19 @@ impl IambWindow {
         }
     }
 
+    pub async fn space_command(
+        &mut self,
+        act: SpaceAction,
+        ctx: ProgramContext,
+        store: &mut ProgramStore,
+    ) -> IambResult<EditInfo> {
+        if let IambWindow::Room(w) = self {
+            w.space_command(act, ctx, store).await
+        } else {
+            return Err(IambError::NoSelectedRoom.into());
+        }
+    }
+
     pub async fn room_command(
         &mut self,
         act: RoomAction,
@@ -496,7 +522,8 @@ impl WindowOps<IambInfo> for IambWindow {
                     .map(|room_info| DirectItem::new(room_info, store))
                     .collect::<Vec<_>>();
                 let fields = &store.application.settings.tunables.sort.dms;
-                items.sort_by(|a, b| room_fields_cmp(a, b, fields));
+                let collator = &mut store.application.collator;
+                items.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
 
                 state.set(items);
 
@@ -541,7 +568,8 @@ impl WindowOps<IambInfo> for IambWindow {
                     .map(|room_info| RoomItem::new(room_info, store))
                     .collect::<Vec<_>>();
                 let fields = &store.application.settings.tunables.sort.rooms;
-                items.sort_by(|a, b| room_fields_cmp(a, b, fields));
+                let collator = &mut store.application.collator;
+                items.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
 
                 state.set(items);
 
@@ -572,7 +600,8 @@ impl WindowOps<IambInfo> for IambWindow {
                 items.extend(dms);
 
                 let fields = &store.application.settings.tunables.sort.chats;
-                items.sort_by(|a, b| room_fields_cmp(a, b, fields));
+                let collator = &mut store.application.collator;
+                items.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
 
                 state.set(items);
 
@@ -605,12 +634,13 @@ impl WindowOps<IambInfo> for IambWindow {
                 items.extend(dms);
 
                 let fields = &store.application.settings.tunables.sort.chats;
-                items.sort_by(|a, b| room_fields_cmp(a, b, fields));
+                let collator = &mut store.application.collator;
+                items.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
 
                 state.set(items);
 
                 List::new(store)
-                    .empty_message("You do not have rooms or dms yet")
+                    .empty_message("You do not have any unreads yet")
                     .empty_alignment(Alignment::Center)
                     .focus(focused)
                     .render(area, buf, state);
@@ -625,7 +655,8 @@ impl WindowOps<IambInfo> for IambWindow {
                     .map(|room| SpaceItem::new(room, store))
                     .collect::<Vec<_>>();
                 let fields = &store.application.settings.tunables.sort.spaces;
-                items.sort_by(|a, b| room_fields_cmp(a, b, fields));
+                let collator = &mut store.application.collator;
+                items.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
 
                 state.set(items);
 
@@ -711,7 +742,7 @@ impl Window<IambInfo> for IambWindow {
         }
     }
 
-    fn get_tab_title(&self, store: &mut ProgramStore) -> Line {
+    fn get_tab_title(&self, store: &mut ProgramStore) -> Line<'_> {
         match self {
             IambWindow::DirectList(_) => bold_spans("Direct Messages"),
             IambWindow::RoomList(_) => bold_spans("Rooms"),
@@ -739,7 +770,7 @@ impl Window<IambInfo> for IambWindow {
         }
     }
 
-    fn get_win_title(&self, store: &mut ProgramStore) -> Line {
+    fn get_win_title(&self, store: &mut ProgramStore) -> Line<'_> {
         match self {
             IambWindow::DirectList(_) => bold_spans("Direct Messages"),
             IambWindow::RoomList(_) => bold_spans("Rooms"),
@@ -769,7 +800,7 @@ impl Window<IambInfo> for IambWindow {
                 let (room, name, tags) = store.application.worker.get_room(room_id)?;
                 let room = RoomState::new(room, thread, name, tags, store);
 
-                store.application.need_load.insert(room.id().to_owned(), Need::MEMBERS);
+                store.application.need_load.need_members(room.id().to_owned());
                 return Ok(room.into());
             },
             IambId::DirectList => {
@@ -831,7 +862,7 @@ impl Window<IambInfo> for IambWindow {
             let (room, name, tags) = store.application.worker.get_room(room_id)?;
             let room = RoomState::new(room, None, name, tags, store);
 
-            store.application.need_load.insert(room.id().to_owned(), Need::MEMBERS);
+            store.application.need_load.need_members(room.id().to_owned());
             Ok(room.into())
         }
     }
@@ -914,6 +945,10 @@ impl RoomLikeItem for GenericChatItem {
     fn is_unread(&self) -> bool {
         self.unread.is_unread()
     }
+
+    fn is_invite(&self) -> bool {
+        self.room().state() == MatrixRoomState::Invited
+    }
 }
 
 impl Display for GenericChatItem {
@@ -923,7 +958,12 @@ impl Display for GenericChatItem {
 }
 
 impl ListItem<IambInfo> for GenericChatItem {
-    fn show(&self, selected: bool, _: &ViewportContext<ListCursor>, _: &mut ProgramStore) -> Text {
+    fn show(
+        &self,
+        selected: bool,
+        _: &ViewportContext<ListCursor>,
+        _: &mut ProgramStore,
+    ) -> Text<'_> {
         let unread = self.unread.is_unread();
         let style = selected_style(selected);
         let (name, mut labels) = name_and_labels(&self.name, unread, style);
@@ -1024,16 +1064,25 @@ impl RoomLikeItem for RoomItem {
     fn is_unread(&self) -> bool {
         self.unread.is_unread()
     }
+
+    fn is_invite(&self) -> bool {
+        self.room().state() == MatrixRoomState::Invited
+    }
 }
 
 impl Display for RoomItem {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, ":verify request {}", self.name)
+        write!(f, "{}", self.name)
     }
 }
 
 impl ListItem<IambInfo> for RoomItem {
-    fn show(&self, selected: bool, _: &ViewportContext<ListCursor>, _: &mut ProgramStore) -> Text {
+    fn show(
+        &self,
+        selected: bool,
+        _: &ViewportContext<ListCursor>,
+        _: &mut ProgramStore,
+    ) -> Text<'_> {
         let unread = self.unread.is_unread();
         let style = selected_style(selected);
         let (name, mut labels) = name_and_labels(&self.name, unread, style);
@@ -1124,6 +1173,10 @@ impl RoomLikeItem for DirectItem {
     fn is_unread(&self) -> bool {
         self.unread.is_unread()
     }
+
+    fn is_invite(&self) -> bool {
+        self.room().state() == MatrixRoomState::Invited
+    }
 }
 
 impl Display for DirectItem {
@@ -1133,7 +1186,12 @@ impl Display for DirectItem {
 }
 
 impl ListItem<IambInfo> for DirectItem {
-    fn show(&self, selected: bool, _: &ViewportContext<ListCursor>, _: &mut ProgramStore) -> Text {
+    fn show(
+        &self,
+        selected: bool,
+        _: &ViewportContext<ListCursor>,
+        _: &mut ProgramStore,
+    ) -> Text<'_> {
         let unread = self.unread.is_unread();
         let style = selected_style(selected);
         let (name, mut labels) = name_and_labels(&self.name, unread, style);
@@ -1223,16 +1281,25 @@ impl RoomLikeItem for SpaceItem {
         // XXX: this needs to check whether the space contains rooms with unread messages
         false
     }
+
+    fn is_invite(&self) -> bool {
+        self.room().state() == MatrixRoomState::Invited
+    }
 }
 
 impl Display for SpaceItem {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, ":verify request {}", self.room_id())
+        write!(f, "{}", self.name)
     }
 }
 
 impl ListItem<IambInfo> for SpaceItem {
-    fn show(&self, selected: bool, _: &ViewportContext<ListCursor>, _: &mut ProgramStore) -> Text {
+    fn show(
+        &self,
+        selected: bool,
+        _: &ViewportContext<ListCursor>,
+        _: &mut ProgramStore,
+    ) -> Text<'_> {
         selected_text(self.name.as_str(), selected)
     }
 
@@ -1363,7 +1430,12 @@ impl Display for VerifyItem {
 }
 
 impl ListItem<IambInfo> for VerifyItem {
-    fn show(&self, selected: bool, _: &ViewportContext<ListCursor>, _: &mut ProgramStore) -> Text {
+    fn show(
+        &self,
+        selected: bool,
+        _: &ViewportContext<ListCursor>,
+        _: &mut ProgramStore,
+    ) -> Text<'_> {
         let mut lines = vec![];
 
         let bold = Style::default().add_modifier(StyleModifier::BOLD);
@@ -1473,7 +1545,7 @@ impl ListItem<IambInfo> for MemberItem {
         selected: bool,
         _: &ViewportContext<ListCursor>,
         store: &mut ProgramStore,
-    ) -> Text {
+    ) -> Text<'_> {
         let info = store.application.rooms.get_or_default(self.room_id.clone());
         let user_id = self.member.user_id();
 
@@ -1517,6 +1589,10 @@ impl ListItem<IambInfo> for MemberItem {
     fn get_word(&self) -> Option<String> {
         self.member.user_id().to_string().into()
     }
+
+    fn matches(&self, needle: &regex::Regex) -> bool {
+        needle.is_match(self.member.name()) || needle.is_match(self.member.user_id().as_str())
+    }
 }
 
 impl Promptable<ProgramContext, ProgramStore, IambInfo> for MemberItem {
@@ -1556,6 +1632,7 @@ mod tests {
         alias: Option<OwnedRoomAliasId>,
         name: &'static str,
         unread: UnreadInfo,
+        invite: bool,
     }
 
     impl RoomLikeItem for &TestRoomItem {
@@ -1582,46 +1659,55 @@ mod tests {
         fn is_unread(&self) -> bool {
             self.unread.is_unread()
         }
+
+        fn is_invite(&self) -> bool {
+            self.invite
+        }
     }
 
     #[test]
     fn test_sort_rooms() {
+        let mut collator = Collator::default();
+        let collator = &mut collator;
         let server = server_name!("example.com");
 
         let room1 = TestRoomItem {
-            room_id: RoomId::new(server).to_owned(),
+            room_id: RoomId::new_v1(server).to_owned(),
             tags: vec![TagName::Favorite],
             alias: Some(room_alias_id!("#room1:example.com").to_owned()),
             name: "Z",
             unread: UnreadInfo::default(),
+            invite: false,
         };
 
         let room2 = TestRoomItem {
-            room_id: RoomId::new(server).to_owned(),
+            room_id: RoomId::new_v1(server).to_owned(),
             tags: vec![],
             alias: Some(room_alias_id!("#a:example.com").to_owned()),
             name: "Unnamed Room",
             unread: UnreadInfo::default(),
+            invite: false,
         };
 
         let room3 = TestRoomItem {
-            room_id: RoomId::new(server).to_owned(),
+            room_id: RoomId::new_v1(server).to_owned(),
             tags: vec![],
             alias: None,
             name: "Cool Room",
             unread: UnreadInfo::default(),
+            invite: false,
         };
 
         // Sort by Name ascending.
         let mut rooms = vec![&room1, &room2, &room3];
         let fields = &[SortColumn(SortFieldRoom::Name, SortOrder::Ascending)];
-        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields));
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
         assert_eq!(rooms, vec![&room3, &room2, &room1]);
 
         // Sort by Name descending.
         let mut rooms = vec![&room1, &room2, &room3];
         let fields = &[SortColumn(SortFieldRoom::Name, SortOrder::Descending)];
-        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields));
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
         assert_eq!(rooms, vec![&room1, &room2, &room3]);
 
         // Sort by Favorite and Alias before Name to show order matters.
@@ -1631,7 +1717,7 @@ mod tests {
             SortColumn(SortFieldRoom::Alias, SortOrder::Ascending),
             SortColumn(SortFieldRoom::Name, SortOrder::Ascending),
         ];
-        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields));
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
         assert_eq!(rooms, vec![&room1, &room2, &room3]);
 
         // Now flip order of Favorite with Descending
@@ -1641,24 +1727,27 @@ mod tests {
             SortColumn(SortFieldRoom::Alias, SortOrder::Ascending),
             SortColumn(SortFieldRoom::Name, SortOrder::Ascending),
         ];
-        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields));
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
         assert_eq!(rooms, vec![&room2, &room3, &room1]);
     }
 
     #[test]
     fn test_sort_room_recents() {
+        let mut collator = Collator::default();
+        let collator = &mut collator;
         let server = server_name!("example.com");
 
         let room1 = TestRoomItem {
-            room_id: RoomId::new(server).to_owned(),
+            room_id: RoomId::new_v1(server).to_owned(),
             tags: vec![],
             alias: None,
             name: "Room 1",
             unread: UnreadInfo { unread: false, latest: None },
+            invite: false,
         };
 
         let room2 = TestRoomItem {
-            room_id: RoomId::new(server).to_owned(),
+            room_id: RoomId::new_v1(server).to_owned(),
             tags: vec![],
             alias: None,
             name: "Room 2",
@@ -1666,10 +1755,11 @@ mod tests {
                 unread: false,
                 latest: Some(MessageTimeStamp::OriginServer(40u32.into())),
             },
+            invite: false,
         };
 
         let room3 = TestRoomItem {
-            room_id: RoomId::new(server).to_owned(),
+            room_id: RoomId::new_v1(server).to_owned(),
             tags: vec![],
             alias: None,
             name: "Room 3",
@@ -1677,18 +1767,71 @@ mod tests {
                 unread: false,
                 latest: Some(MessageTimeStamp::OriginServer(20u32.into())),
             },
+            invite: false,
         };
 
         // Sort by Recent ascending.
         let mut rooms = vec![&room1, &room2, &room3];
         let fields = &[SortColumn(SortFieldRoom::Recent, SortOrder::Ascending)];
-        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields));
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
         assert_eq!(rooms, vec![&room2, &room3, &room1]);
 
         // Sort by Recent descending.
         let mut rooms = vec![&room1, &room2, &room3];
         let fields = &[SortColumn(SortFieldRoom::Recent, SortOrder::Descending)];
-        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields));
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
         assert_eq!(rooms, vec![&room1, &room3, &room2]);
     }
+
+    #[test]
+    fn test_sort_room_invites() {
+        let mut collator = Collator::default();
+        let collator = &mut collator;
+        let server = server_name!("example.com");
+
+        let room1 = TestRoomItem {
+            room_id: RoomId::new_v1(server).to_owned(),
+            tags: vec![],
+            alias: None,
+            name: "Old room 1",
+            unread: UnreadInfo::default(),
+            invite: false,
+        };
+
+        let room2 = TestRoomItem {
+            room_id: RoomId::new_v1(server).to_owned(),
+            tags: vec![],
+            alias: None,
+            name: "Old room 2",
+            unread: UnreadInfo::default(),
+            invite: false,
+        };
+
+        let room3 = TestRoomItem {
+            room_id: RoomId::new_v1(server).to_owned(),
+            tags: vec![],
+            alias: None,
+            name: "New Fancy Room",
+            unread: UnreadInfo::default(),
+            invite: true,
+        };
+
+        // Sort invites first
+        let mut rooms = vec![&room1, &room2, &room3];
+        let fields = &[
+            SortColumn(SortFieldRoom::Invite, SortOrder::Ascending),
+            SortColumn(SortFieldRoom::Name, SortOrder::Ascending),
+        ];
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
+        assert_eq!(rooms, vec![&room3, &room1, &room2]);
+
+        // Sort invites after
+        let mut rooms = vec![&room1, &room2, &room3];
+        let fields = &[
+            SortColumn(SortFieldRoom::Invite, SortOrder::Descending),
+            SortColumn(SortFieldRoom::Name, SortOrder::Ascending),
+        ];
+        rooms.sort_by(|a, b| room_fields_cmp(a, b, fields, collator));
+        assert_eq!(rooms, vec![&room1, &room2, &room3]);
+    }
 }
