@@ -7,14 +7,16 @@ use std::path::{Path, PathBuf};
 
 use edit::edit_with_builder as external_edit;
 use edit::Builder;
+use matrix_sdk::EncryptionState;
 use modalkit::editing::store::RegisterError;
+use ratatui::style::{Color, Style};
 use std::process::Command;
 use tokio;
 use url::Url;
 
 use matrix_sdk::{
     attachment::AttachmentConfig,
-    media::{MediaFormat, MediaRequest},
+    media::{MediaFormat, MediaRequestParameters},
     room::Room as MatrixRoom,
     ruma::{
         events::reaction::ReactionEventContent,
@@ -86,7 +88,14 @@ use crate::base::{
     SendAction,
 };
 
-use crate::message::{text_to_message, Message, MessageEvent, MessageKey, MessageTimeStamp};
+use crate::message::{
+    text_to_message,
+    Message,
+    MessageEvent,
+    MessageKey,
+    MessageTimeStamp,
+    TreeGenState,
+};
 use crate::worker::Requester;
 
 use super::scrollback::{Scrollback, ScrollbackState};
@@ -213,12 +222,10 @@ impl ChatState {
                     };
 
                     let (source, msg_filename) = match &ev.content.msgtype {
-                        MessageType::Audio(c) => (c.source.clone(), c.body.as_str()),
-                        MessageType::File(c) => {
-                            (c.source.clone(), c.filename.as_deref().unwrap_or(c.body.as_str()))
-                        },
-                        MessageType::Image(c) => (c.source.clone(), c.body.as_str()),
-                        MessageType::Video(c) => (c.source.clone(), c.body.as_str()),
+                        MessageType::Audio(c) => (c.source.clone(), c.filename()),
+                        MessageType::File(c) => (c.source.clone(), c.filename()),
+                        MessageType::Image(c) => (c.source.clone(), c.filename()),
+                        MessageType::Video(c) => (c.source.clone(), c.filename()),
                         _ => {
                             if !flags.contains(DownloadFlags::OPEN) {
                                 return Err(IambError::NoAttachment.into());
@@ -226,10 +233,14 @@ impl ChatState {
 
                             let links = if let Some(html) = &msg.html {
                                 html.get_links()
-                            } else if let Ok(url) = Url::parse(&msg.event.body()) {
-                                vec![('0', url)]
                             } else {
-                                vec![]
+                                linkify::LinkFinder::new()
+                                    .links(&msg.event.body())
+                                    .filter_map(|u| Url::parse(u.as_str()).ok())
+                                    .scan(TreeGenState { link_num: 0 }, |state, u| {
+                                        state.next_link_char().map(|c| (c, u))
+                                    })
+                                    .collect()
                             };
 
                             if links.is_empty() {
@@ -252,7 +263,7 @@ impl ChatState {
                     };
 
                     if filename.is_dir() {
-                        filename.push(msg_filename);
+                        filename.push(msg_filename.replace(std::path::MAIN_SEPARATOR_STR, "_"));
                     }
 
                     if filename.exists() && !flags.contains(DownloadFlags::FORCE) {
@@ -262,9 +273,9 @@ impl ChatState {
                             let mut filename_incr = filename.clone();
                             for n in 1..=1000 {
                                 if let Some(ext) = ext.and_then(OsStr::to_str) {
-                                    filename_incr.set_file_name(format!("{}-{}.{}", stem, n, ext));
+                                    filename_incr.set_file_name(format!("{stem}-{n}.{ext}"));
                                 } else {
-                                    filename_incr.set_file_name(format!("{}-{}", stem, n));
+                                    filename_incr.set_file_name(format!("{stem}-{n}"));
                                 }
 
                                 if !filename_incr.exists() {
@@ -276,7 +287,7 @@ impl ChatState {
                     }
 
                     if !filename.exists() || flags.contains(DownloadFlags::FORCE) {
-                        let req = MediaRequest { source, format: MediaFormat::File };
+                        let req = MediaRequestParameters { source, format: MediaFormat::File };
 
                         let bytes =
                             media.get_media_content(&req, true).await.map_err(IambError::from)?;
@@ -380,6 +391,7 @@ impl ChatState {
                     MessageEvent::EncryptedRedacted(ev) => ev.event_id.clone(),
                     MessageEvent::Original(ev) => ev.event_id.clone(),
                     MessageEvent::Local(event_id, _) => event_id.clone(),
+                    MessageEvent::State(ev) => ev.event_id().to_owned(),
                     MessageEvent::Redacted(_) => {
                         let msg = "Cannot react to a redacted message";
                         let err = UIError::Failure(msg.into());
@@ -389,7 +401,7 @@ impl ChatState {
                 };
 
                 if info.user_reactions_contains(&settings.profile.user_id, &event_id, &emoji) {
-                    let msg = format!("You’ve already reacted to this message with {}", emoji);
+                    let msg = format!("You’ve already reacted to this message with {emoji}");
                     let err = UIError::Failure(msg);
 
                     return Err(err);
@@ -417,6 +429,7 @@ impl ChatState {
                     MessageEvent::EncryptedRedacted(ev) => ev.event_id.clone(),
                     MessageEvent::Original(ev) => ev.event_id.clone(),
                     MessageEvent::Local(event_id, _) => event_id.clone(),
+                    MessageEvent::State(ev) => ev.event_id().to_owned(),
                     MessageEvent::Redacted(_) => {
                         let msg = "Cannot redact already redacted message";
                         let err = UIError::Failure(msg.into());
@@ -437,6 +450,21 @@ impl ChatState {
 
                 Ok(None)
             },
+            MessageAction::Replied => {
+                let Some(reply) = msg.reply_to() else {
+                    let msg = "Selected message is not a reply";
+                    return Err(UIError::Failure(msg.into()));
+                };
+
+                let Some(key) = info.get_message_key(&reply) else {
+                    store.application.need_load.need_message(self.room_id.clone(), reply);
+                    let msg = "Replied to message will be loaded in the background";
+                    return Err(UIError::Failure(msg.into()));
+                };
+
+                self.scrollback.goto_message(key.clone());
+                Ok(None)
+            },
             MessageAction::Unreact(reaction, literal) => {
                 let emoji = match reaction {
                     reaction if literal => reaction,
@@ -464,6 +492,7 @@ impl ChatState {
                     MessageEvent::EncryptedRedacted(ev) => ev.event_id.clone(),
                     MessageEvent::Original(ev) => ev.event_id.clone(),
                     MessageEvent::Local(event_id, _) => event_id.clone(),
+                    MessageEvent::State(ev) => ev.event_id().to_owned(),
                     MessageEvent::Redacted(_) => {
                         let msg = "Cannot unreact to a redacted message";
                         let err = UIError::Failure(msg.into());
@@ -596,17 +625,16 @@ impl ChatState {
                             let dynimage = image::DynamicImage::ImageRgba8(imagebuf);
                             let bytes = Vec::<u8>::new();
                             let mut buff = std::io::Cursor::new(bytes);
-                            dynimage.write_to(&mut buff, image::ImageOutputFormat::Png)?;
+                            dynimage.write_to(&mut buff, image::ImageFormat::Png)?;
                             Ok(buff.into_inner())
-                        })
-                        .map_err(IambError::from)?;
+                        })?;
                 let mime = mime::IMAGE_PNG;
 
                 let name = "Clipboard.png";
                 let config = AttachmentConfig::new();
 
                 let resp = room
-                    .send_attachment(name.as_ref(), &mime, bytes, config)
+                    .send_attachment(name, &mime, bytes, config)
                     .await
                     .map_err(IambError::from)?;
 
@@ -635,10 +663,7 @@ impl ChatState {
     }
 
     pub fn focus_toggle(&mut self) {
-        self.focus = match self.focus {
-            RoomFocus::Scrollback => RoomFocus::MessageBar,
-            RoomFocus::MessageBar => RoomFocus::Scrollback,
-        };
+        self.focus.toggle();
     }
 
     pub fn room(&self) -> &MatrixRoom {
@@ -649,6 +674,14 @@ impl ChatState {
         &self.room_id
     }
 
+    pub fn auto_toggle_focus(
+        &mut self,
+        act: &EditorAction,
+        ctx: &ProgramContext,
+    ) -> Option<EditorAction> {
+        auto_toggle_focus(&mut self.focus, act, ctx, &self.scrollback, &mut self.tbox)
+    }
+
     pub fn typing_notice(
         &self,
         act: &EditorAction,
@@ -751,8 +784,15 @@ impl Editable<ProgramContext, ProgramStore, IambInfo> for ChatState {
         ctx: &ProgramContext,
         store: &mut ProgramStore,
     ) -> EditResult<EditInfo, IambInfo> {
+        // Check whether we should automatically switch between the message bar
+        // or message scrollback, and use an adjusted action if we do so.
+        let adjusted = self.auto_toggle_focus(act, ctx);
+        let act = adjusted.as_ref().unwrap_or(act);
+
+        // Send typing notice if needed.
         self.typing_notice(act, ctx, store);
 
+        // And now we can finally run the editor command.
         match delegate!(self, w => w.editor_command(act, ctx, store)) {
             res @ Ok(_) => res,
             Err(EditError::WrongBuffer(IambBufferId::Room(room_id, thread, focus)))
@@ -849,16 +889,16 @@ impl PromptActions<ProgramContext, ProgramStore, IambInfo> for ChatState {
 
     fn recall(
         &mut self,
+        filter: &RecallFilter,
         dir: &MoveDir1D,
         count: &Count,
-        prefixed: bool,
         ctx: &ProgramContext,
         _: &mut ProgramStore,
     ) -> EditResult<Vec<(ProgramAction, ProgramContext)>, IambInfo> {
         let count = ctx.resolve(count);
         let rope = self.tbox.get();
 
-        let text = self.sent.recall(&rope, &mut self.sent_scrollback, *dir, prefixed, count);
+        let text = self.sent.recall(&rope, &mut self.sent_scrollback, filter, *dir, count);
 
         if let Some(text) = text {
             self.tbox.set_text(text);
@@ -882,9 +922,7 @@ impl Promptable<ProgramContext, ProgramStore, IambInfo> for ChatState {
         match act {
             PromptAction::Submit => self.submit(ctx, store),
             PromptAction::Abort(empty) => self.abort(*empty, ctx, store),
-            PromptAction::Recall(dir, count, prefixed) => {
-                self.recall(dir, count, *prefixed, ctx, store)
-            },
+            PromptAction::Recall(filter, dir, count) => self.recall(filter, dir, count, ctx, store),
         }
     }
 }
@@ -906,7 +944,7 @@ impl<'a> Chat<'a> {
     }
 }
 
-impl<'a> StatefulWidget for Chat<'a> {
+impl StatefulWidget for Chat<'_> {
     type State = ChatState;
 
     fn render(self, area: Rect, buf: &mut Buffer, state: &mut Self::State) {
@@ -954,7 +992,16 @@ impl<'a> StatefulWidget for Chat<'a> {
             Paragraph::new(desc_spans).render(descarea, buf);
         }
 
-        let prompt = if self.focused { "> " } else { "  " };
+        let prompt = match (self.focused, state.room().encryption_state()) {
+            (false, _) => Span::raw("  "),
+            (_, EncryptionState::Encrypted) => {
+                Span::styled("\u{1F512}\u{FE0E} ", Style::new().fg(Color::LightGreen))
+            },
+            (_, EncryptionState::NotEncrypted) => {
+                Span::styled("\u{1F513}\u{FE0E} ", Style::new().fg(Color::Red))
+            },
+            (_, EncryptionState::Unknown) => Span::styled("> ", Style::new().fg(Color::Red)),
+        };
 
         let tbox = TextBox::new().prompt(prompt);
         tbox.render(textarea, buf, &mut state.tbox);
@@ -990,3 +1037,158 @@ fn cmd(open_command: &Vec<String>) -> Option<Command> {
     }
     None
 }
+
+pub fn auto_toggle_focus(
+    focus: &mut RoomFocus,
+    act: &EditorAction,
+    ctx: &ProgramContext,
+    scrollback: &ScrollbackState,
+    tbox: &mut TextBoxState<IambInfo>,
+) -> Option<EditorAction> {
+    let is_insert = ctx.get_insert_style().is_some();
+
+    match (focus, act) {
+        (f @ RoomFocus::Scrollback, _) if is_insert => {
+            // Insert mode commands should switch focus.
+            f.toggle();
+            None
+        },
+        (f @ RoomFocus::Scrollback, EditorAction::InsertText(_)) => {
+            // Pasting or otherwise inserting text should switch.
+            f.toggle();
+            None
+        },
+        (
+            f @ RoomFocus::Scrollback,
+            EditorAction::Edit(
+                op,
+                EditTarget::Motion(mov @ MoveType::Line(MoveDir1D::Next), count),
+            ),
+        ) if ctx.resolve(op).is_motion() => {
+            let count = ctx.resolve(count);
+
+            if count > 0 && scrollback.is_latest() {
+                // Trying to move down a line when already at the end of room history should
+                // switch.
+                f.toggle();
+
+                // And decrement the count for the action.
+                let count = count.saturating_sub(1).into();
+                let target = EditTarget::Motion(mov.clone(), count);
+                let dec = EditorAction::Edit(op.clone(), target);
+
+                Some(dec)
+            } else {
+                None
+            }
+        },
+        (
+            f @ RoomFocus::MessageBar,
+            EditorAction::Edit(
+                op,
+                EditTarget::Motion(mov @ MoveType::Line(MoveDir1D::Previous), count),
+            ),
+        ) if !is_insert && ctx.resolve(op).is_motion() => {
+            let count = ctx.resolve(count);
+
+            if count > 0 && tbox.get_cursor().y == 0 {
+                // Trying to move up a line when already at the top of the msgbar should
+                // switch as long as we're not in Insert mode.
+                f.toggle();
+
+                // And decrement the count for the action.
+                let count = count.saturating_sub(1).into();
+                let target = EditTarget::Motion(mov.clone(), count);
+                let dec = EditorAction::Edit(op.clone(), target);
+
+                Some(dec)
+            } else {
+                None
+            }
+        },
+        (RoomFocus::Scrollback, _) | (RoomFocus::MessageBar, _) => {
+            // Do not switch.
+            None
+        },
+    }
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    use modalkit::actions::{EditAction, InsertTextAction};
+
+    use crate::tests::{mock_store, TEST_ROOM1_ID};
+
+    macro_rules! move_line {
+        ($dir: expr, $count: expr) => {
+            EditorAction::Edit(
+                EditAction::Motion.into(),
+                EditTarget::Motion(MoveType::Line($dir), $count.into()),
+            )
+        };
+    }
+
+    #[tokio::test]
+    async fn test_auto_focus() {
+        let mut store = mock_store().await;
+        let ctx = ProgramContext::default();
+
+        let room_id = TEST_ROOM1_ID.clone();
+        let scrollback = ScrollbackState::new(room_id.clone(), None);
+
+        let id = IambBufferId::Room(room_id, None, RoomFocus::MessageBar);
+        let ebuf = store.load_buffer(id);
+        let mut tbox = TextBoxState::new(ebuf);
+
+        // Start out focused on the scrollback.
+        let mut focused = RoomFocus::Scrollback;
+
+        // Inserting text toggles:
+        let act = EditorAction::InsertText(InsertTextAction::Type(
+            Char::from('a').into(),
+            MoveDir1D::Next,
+            1.into(),
+        ));
+        let res = auto_toggle_focus(&mut focused, &act, &ctx, &scrollback, &mut tbox);
+        assert_eq!(focused, RoomFocus::MessageBar);
+        assert!(res.is_none());
+
+        // Going down in message bar doesn't toggle:
+        let act = move_line!(MoveDir1D::Next, 1);
+        let res = auto_toggle_focus(&mut focused, &act, &ctx, &scrollback, &mut tbox);
+        assert_eq!(focused, RoomFocus::MessageBar);
+        assert!(res.is_none());
+
+        // But going up will:
+        let act = move_line!(MoveDir1D::Previous, 1);
+        let res = auto_toggle_focus(&mut focused, &act, &ctx, &scrollback, &mut tbox);
+        assert_eq!(focused, RoomFocus::Scrollback);
+        assert_eq!(res, Some(move_line!(MoveDir1D::Previous, 0)));
+
+        // Going up in scrollback doesn't toggle:
+        let act = move_line!(MoveDir1D::Previous, 1);
+        let res = auto_toggle_focus(&mut focused, &act, &ctx, &scrollback, &mut tbox);
+        assert_eq!(focused, RoomFocus::Scrollback);
+        assert_eq!(res, None);
+
+        // And then go back down:
+        let act = move_line!(MoveDir1D::Next, 1);
+        let res = auto_toggle_focus(&mut focused, &act, &ctx, &scrollback, &mut tbox);
+        assert_eq!(focused, RoomFocus::MessageBar);
+        assert_eq!(res, Some(move_line!(MoveDir1D::Next, 0)));
+
+        // Go up 2 will go up 1 in scrollback:
+        let act = move_line!(MoveDir1D::Previous, 2);
+        let res = auto_toggle_focus(&mut focused, &act, &ctx, &scrollback, &mut tbox);
+        assert_eq!(focused, RoomFocus::Scrollback);
+        assert_eq!(res, Some(move_line!(MoveDir1D::Previous, 1)));
+
+        // Go down 3 will go down 2 in messagebar:
+        let act = move_line!(MoveDir1D::Next, 3);
+        let res = auto_toggle_focus(&mut focused, &act, &ctx, &scrollback, &mut tbox);
+        assert_eq!(focused, RoomFocus::MessageBar);
+        assert_eq!(res, Some(move_line!(MoveDir1D::Next, 2)));
+    }
+}
