@@ -26,7 +26,7 @@ use matrix_sdk::{
         OwnedUserId,
         RoomId,
     },
-    DisplayName,
+    RoomDisplayName,
     RoomState as MatrixRoomState,
 };
 
@@ -66,6 +66,7 @@ use crate::base::{
     RoomAction,
     RoomField,
     SendAction,
+    SpaceAction,
 };
 
 use self::chat::ChatState;
@@ -119,19 +120,19 @@ fn hist_visibility_mode(name: impl Into<String>) -> IambResult<HistoryVisibility
 /// that operations like sending and accepting invites, opening the members window, etc., all work
 /// similarly.
 pub enum RoomState {
-    Chat(ChatState),
-    Space(SpaceState),
+    Chat(Box<ChatState>),
+    Space(Box<SpaceState>),
 }
 
 impl From<ChatState> for RoomState {
     fn from(chat: ChatState) -> Self {
-        RoomState::Chat(chat)
+        RoomState::Chat(Box::new(chat))
     }
 }
 
 impl From<SpaceState> for RoomState {
     fn from(space: SpaceState) -> Self {
-        RoomState::Space(space)
+        RoomState::Space(Box::new(space))
     }
 }
 
@@ -139,7 +140,7 @@ impl RoomState {
     pub fn new(
         room: MatrixRoom,
         thread: Option<OwnedEventId>,
-        name: DisplayName,
+        name: RoomDisplayName,
         tags: Option<Tags>,
         store: &mut ProgramStore,
     ) -> Self {
@@ -214,6 +215,18 @@ impl RoomState {
         }
     }
 
+    pub async fn space_command(
+        &mut self,
+        act: SpaceAction,
+        ctx: ProgramContext,
+        store: &mut ProgramStore,
+    ) -> IambResult<EditInfo> {
+        match self {
+            RoomState::Space(space) => space.space_command(act, ctx, store).await,
+            RoomState::Chat(_) => Err(IambError::NoSelectedSpace.into()),
+        }
+    }
+
     pub async fn send_command(
         &mut self,
         act: SendAction,
@@ -406,7 +419,7 @@ impl RoomState {
                         // Try creating the room alias on the server.
                         let alias_create_req =
                             CreateAliasRequest::new(orai.clone(), room.room_id().into());
-                        if let Err(e) = client.send(alias_create_req, None).await {
+                        if let Err(e) = client.send(alias_create_req).await {
                             if let Some(ClientApiErrorKind::Unknown) = e.client_api_error_kind() {
                                 // Ignore when it already exists.
                             } else {
@@ -447,7 +460,7 @@ impl RoomState {
 
                         // If the room alias does not exist on the server, create it
                         let alias_create_req = CreateAliasRequest::new(orai, room.room_id().into());
-                        if let Err(e) = client.send(alias_create_req, None).await {
+                        if let Err(e) = client.send(alias_create_req).await {
                             if let Some(ClientApiErrorKind::Unknown) = e.client_api_error_kind() {
                                 // Ignore when it already exists.
                             } else {
@@ -464,6 +477,9 @@ impl RoomState {
                     RoomField::Aliases => {
                         // This never happens, aliases is only used for showing
                     },
+                    RoomField::Id => {
+                        // This never happens, id is only used for showing
+                    },
                 }
 
                 Ok(vec![])
@@ -519,7 +535,7 @@ impl RoomState {
                             .application
                             .worker
                             .client
-                            .send(del_req, None)
+                            .send(del_req)
                             .await
                             .map_err(IambError::from)?;
                     },
@@ -552,13 +568,16 @@ impl RoomState {
                             .application
                             .worker
                             .client
-                            .send(del_req, None)
+                            .send(del_req)
                             .await
                             .map_err(IambError::from)?;
                     },
                     RoomField::Aliases => {
                         // This will not happen, you cannot unset all aliases
                     },
+                    RoomField::Id => {
+                        // This never happens, id is only used for showing
+                    },
                 }
 
                 Ok(vec![])
@@ -572,7 +591,12 @@ impl RoomState {
                 let msg = match field {
                     RoomField::History => {
                         let visibility = room.history_visibility();
-                        format!("Room history visibility: {visibility}")
+                        let visibility = visibility.as_ref().map(|v| v.as_str());
+                        format!("Room history visibility: {}", visibility.unwrap_or("<unknown>"))
+                    },
+                    RoomField::Id => {
+                        let id = room.room_id();
+                        format!("Room identifier: {id}")
                     },
                     RoomField::Name => {
                         match room.name() {
@@ -634,7 +658,7 @@ impl RoomState {
         }
     }
 
-    pub fn get_title(&self, store: &mut ProgramStore) -> Line {
+    pub fn get_title(&self, store: &mut ProgramStore) -> Line<'_> {
         let title = store.application.get_room_title(self.id());
         let style = Style::default().add_modifier(StyleModifier::BOLD);
         let mut spans = vec![];
@@ -752,8 +776,8 @@ impl WindowOps<IambInfo> for RoomState {
 
     fn dup(&self, store: &mut ProgramStore) -> Self {
         match self {
-            RoomState::Chat(chat) => RoomState::Chat(chat.dup(store)),
-            RoomState::Space(space) => RoomState::Space(space.dup(store)),
+            RoomState::Chat(chat) => RoomState::Chat(Box::new(chat.dup(store))),
+            RoomState::Space(space) => RoomState::Space(Box::new(space.dup(store))),
         }
     }
 
