@@ -47,7 +47,6 @@ use crate::{
         IambId,
         IambInfo,
         IambResult,
-        Need,
         ProgramContext,
         ProgramStore,
         RoomFetchStatus,
@@ -79,14 +78,20 @@ fn nth_key_before(pos: MessageKey, n: usize, thread: &Messages) -> MessageKey {
 }
 
 fn nth_before(pos: MessageKey, n: usize, thread: &Messages) -> MessageCursor {
-    nth_key_before(pos, n, thread).into()
+    let key = nth_key_before(pos, n, thread);
+
+    if matches!(thread.last_key_value(), Some((last, _)) if &key == last) {
+        MessageCursor::latest()
+    } else {
+        MessageCursor::from(key)
+    }
 }
 
-fn nth_key_after(pos: MessageKey, n: usize, thread: &Messages) -> MessageKey {
+fn nth_key_after(pos: MessageKey, n: usize, thread: &Messages) -> Option<MessageKey> {
     let mut end = &pos;
-    let iter = thread.range(&pos..).enumerate();
+    let mut iter = thread.range(&pos..).enumerate();
 
-    for (i, (key, _)) in iter {
+    for (i, (key, _)) in iter.by_ref() {
         end = key;
 
         if i >= n {
@@ -94,11 +99,12 @@ fn nth_key_after(pos: MessageKey, n: usize, thread: &Messages) -> MessageKey {
         }
     }
 
-    end.clone()
+    // Avoid returning the key if it's at the end.
+    iter.next().map(|_| end.clone())
 }
 
 fn nth_after(pos: MessageKey, n: usize, thread: &Messages) -> MessageCursor {
-    nth_key_after(pos, n, thread).into()
+    nth_key_after(pos, n, thread).map(MessageCursor::from).unwrap_or_default()
 }
 
 fn prevmsg<'a>(key: &MessageKey, thread: &'a Messages) -> Option<&'a Message> {
@@ -150,10 +156,20 @@ impl ScrollbackState {
         }
     }
 
+    pub fn is_latest(&self) -> bool {
+        self.cursor.timestamp.is_none()
+    }
+
     pub fn goto_latest(&mut self) {
         self.cursor = MessageCursor::latest();
     }
 
+    pub fn goto_message(&mut self, target: MessageKey) {
+        let mut cursor = MessageCursor::new(target, 0);
+        std::mem::swap(&mut cursor, &mut self.cursor);
+        self.jumped.push(cursor);
+    }
+
     /// Set the dimensions and placement within the terminal window for this list.
     pub fn set_term_info(&mut self, area: Rect) {
         self.viewctx.dimensions = (area.width as usize, area.height as usize);
@@ -678,10 +694,7 @@ impl EditorActions<ProgramContext, ProgramStore, IambInfo> for ScrollbackState {
 
                         let (mc, needs_load) = self.find_message(key, dir, &needle, count, info);
                         if needs_load {
-                            store
-                                .application
-                                .need_load
-                                .insert(self.room_id.clone(), Need::MESSAGES);
+                            store.application.need_load.need_messages(self.room_id.clone());
                         }
                         mc
                     },
@@ -757,10 +770,7 @@ impl EditorActions<ProgramContext, ProgramStore, IambInfo> for ScrollbackState {
 
                         let (mc, needs_load) = self.find_message(key, dir, &needle, count, info);
                         if needs_load {
-                            store
-                                .application
-                                .need_load
-                                .insert(self.room_id.to_owned(), Need::MESSAGES);
+                            store.application.need_load.need_messages(self.room_id.to_owned());
                         }
 
                         mc.map(|c| self._range_to(c))
@@ -829,8 +839,8 @@ impl EditorActions<ProgramContext, ProgramStore, IambInfo> for ScrollbackState {
 
     fn complete(
         &mut self,
+        _: &CompletionStyle,
         _: &CompletionType,
-        _: &CompletionSelection,
         _: &CompletionDisplay,
         _: &ProgramContext,
         _: &mut ProgramStore,
@@ -1284,7 +1294,7 @@ impl<'a> Scrollback<'a> {
     }
 }
 
-impl<'a> StatefulWidget for Scrollback<'a> {
+impl StatefulWidget for Scrollback<'_> {
     type State = ScrollbackState;
 
     fn render(self, area: Rect, buf: &mut Buffer, state: &mut Self::State) {
@@ -1317,10 +1327,7 @@ impl<'a> StatefulWidget for Scrollback<'a> {
             k
         } else {
             if state.need_more_messages(info) {
-                self.store
-                    .application
-                    .need_load
-                    .insert(state.room_id.to_owned(), Need::MESSAGES);
+                self.store.application.need_load.need_messages(state.room_id.to_owned());
             }
             return;
         };
@@ -1340,7 +1347,7 @@ impl<'a> StatefulWidget for Scrollback<'a> {
 
         for (key, item) in thread.range(&corner_key..) {
             let sel = key == cursor_key;
-            let (txt, mut msg_preview) =
+            let (txt, [mut msg_preview, mut reply_preview]) =
                 item.show_with_preview(prev, foc && sel, &state.viewctx, info, settings);
 
             let incomplete_ok = !full || !sel;
@@ -1357,11 +1364,17 @@ impl<'a> StatefulWidget for Scrollback<'a> {
                     continue;
                 }
 
+                // Only take the preview into the matching row number.
+                // `reply` and `msg` previews are on rows,
+                // so an `or` works to pick the one that matches (if any)
                 let line_preview = match msg_preview {
-                    // Only take the preview into the matching row number.
                     Some((_, _, y)) if y as usize == row => msg_preview.take(),
                     _ => None,
-                };
+                }
+                .or(match reply_preview {
+                    Some((_, _, y)) if y as usize == row => reply_preview.take(),
+                    _ => None,
+                });
 
                 lines.push((key, row, line, line_preview));
                 sawit |= sel;
@@ -1396,7 +1409,7 @@ impl<'a> StatefulWidget for Scrollback<'a> {
         // line.
         for (x, y, backend) in image_previews {
             let image_widget = Image::new(backend);
-            let mut rect = backend.rect();
+            let mut rect = backend.area();
             rect.x = x;
             rect.y = y;
             // Don't render outside of scrollback area
@@ -1411,17 +1424,14 @@ impl<'a> StatefulWidget for Scrollback<'a> {
         {
             // If the cursor is at the last message, then update the read marker.
             if let Some((k, _)) = thread.last_key_value() {
-                info.set_receipt(settings.profile.user_id.clone(), k.1.clone());
+                info.set_receipt(thread.1.clone(), settings.profile.user_id.clone(), k.1.clone());
             }
         }
 
         // Check whether we should load older messages for this room.
         if state.need_more_messages(info) {
             // If the top of the screen is the older message, load more.
-            self.store
-                .application
-                .need_load
-                .insert(state.room_id.to_owned(), Need::MESSAGES);
+            self.store.application.need_load.need_messages(state.room_id.to_owned());
         }
 
         info.draw_last = self.store.application.draw_curr;
@@ -1431,7 +1441,7 @@ impl<'a> StatefulWidget for Scrollback<'a> {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::tests::*;
+    use crate::{base::Need, tests::*};
 
     #[tokio::test]
     async fn test_search_messages() {
@@ -1476,7 +1486,7 @@ mod tests {
             std::mem::take(&mut store.application.need_load)
                 .into_iter()
                 .collect::<Vec<(OwnedRoomId, Need)>>(),
-            vec![(room_id.clone(), Need::MESSAGES)]
+            vec![(room_id.clone(), Need { messages: Some(Vec::new()), members: false })]
         );
 
         // Search forward twice to MSG1.
@@ -1518,8 +1528,9 @@ mod tests {
         scrollback.edit(&EditAction::Motion, &next(1), &ctx, &mut store).unwrap();
         assert_eq!(scrollback.cursor, MSG5_KEY.clone().into());
 
+        // And one more becomes "latest" cursor:
         scrollback.edit(&EditAction::Motion, &next(1), &ctx, &mut store).unwrap();
-        assert_eq!(scrollback.cursor, MSG1_KEY.clone().into());
+        assert_eq!(scrollback.cursor, MessageCursor::latest());
     }
 
     #[tokio::test]
@@ -1553,7 +1564,7 @@ mod tests {
         // MSG1: |                   XXXday, Month NN 20XX                    |
         //       |           @user1:example.com  writhe                       |
         //       |------------------------------------------------------------|
-        let area = Rect::new(0, 0, 60, 4);
+        let area = Rect::new(0, 0, 60, 5);
         let mut buffer = Buffer::empty(area);
         scrollback.draw(area, &mut buffer, true, &mut store);
 
