@@ -20,11 +20,12 @@ use tracing::{error, warn};
 use url::Url;
 
 use matrix_sdk::{
+    authentication::matrix::MatrixSession,
     config::{RequestConfig, SyncSettings},
+    deserialized_responses::DisplayName,
     encryption::verification::{SasVerification, Verification},
     encryption::{BackupDownloadStrategy, EncryptionSettings},
     event_handler::Ctx,
-    matrix_auth::MatrixSession,
     reqwest,
     room::{Messages, MessagesOptions, Room as MatrixRoom, RoomMember},
     ruma::{
@@ -58,6 +59,7 @@ use matrix_sdk::{
             typing::SyncTypingEvent,
             AnyInitialStateEvent,
             AnyMessageLikeEvent,
+            AnySyncStateEvent,
             AnyTimelineEvent,
             EmptyStateKey,
             InitialStateEvent,
@@ -78,15 +80,15 @@ use matrix_sdk::{
     },
     Client,
     ClientBuildError,
-    DisplayName,
     Error as MatrixError,
+    RoomDisplayName,
     RoomMemberships,
 };
 
 use modalkit::errors::UIError;
 use modalkit::prelude::{EditInfo, InfoMessage};
 
-use crate::base::Need;
+use crate::base::MessageNeed;
 use crate::notifications::register_notifications;
 use crate::{
     base::{
@@ -114,8 +116,7 @@ const IAMB_DEVICE_NAME: &str = "iamb";
 const IAMB_USER_AGENT: &str = "iamb";
 const MIN_MSG_LOAD: u32 = 50;
 
-type MessageFetchResult =
-    IambResult<(Option<String>, Vec<(AnyMessageLikeEvent, Vec<OwnedUserId>)>)>;
+type MessageFetchResult = IambResult<(Option<String>, Vec<(AnyTimelineEvent, Vec<OwnedUserId>)>)>;
 
 fn initial_devname() -> String {
     format!("{} on {}", IAMB_DEVICE_NAME, gethostname().to_string_lossy())
@@ -209,13 +210,13 @@ async fn update_event_receipts(info: &mut RoomInfo, room: &MatrixRoom, event_id:
     };
 
     for (user_id, _) in receipts {
-        info.set_receipt(user_id, event_id.to_owned());
+        info.set_receipt(ReceiptThread::Main, user_id, event_id.to_owned());
     }
 }
 
 #[derive(Debug)]
 enum Plan {
-    Messages(OwnedRoomId, Option<String>),
+    Messages(OwnedRoomId, Option<String>, Vec<MessageNeed>),
     Members(OwnedRoomId),
 }
 
@@ -224,8 +225,8 @@ async fn load_plans(store: &AsyncProgramStore) -> Vec<Plan> {
     let ChatStore { need_load, rooms, .. } = &mut locked.application;
     let mut plan = Vec::with_capacity(need_load.rooms() * 2);
 
-    for (room_id, mut need) in std::mem::take(need_load).into_iter() {
-        if need.contains(Need::MESSAGES) {
+    for (room_id, need) in std::mem::take(need_load).into_iter() {
+        if let Some(message_need) = need.messages {
             let info = rooms.get_or_default(room_id.clone());
 
             if !info.recently_fetched() && !info.fetching {
@@ -238,16 +239,11 @@ async fn load_plans(store: &AsyncProgramStore) -> Vec<Plan> {
                     RoomFetchStatus::NotStarted => None,
                 };
 
-                plan.push(Plan::Messages(room_id.to_owned(), fetch_id));
-                need.remove(Need::MESSAGES);
+                plan.push(Plan::Messages(room_id.to_owned(), fetch_id, message_need));
             }
         }
-        if need.contains(Need::MEMBERS) {
+        if need.members {
             plan.push(Plan::Members(room_id.to_owned()));
-            need.remove(Need::MEMBERS);
-        }
-        if !need.is_empty() {
-            need_load.insert(room_id, need);
         }
     }
 
@@ -257,14 +253,14 @@ async fn load_plans(store: &AsyncProgramStore) -> Vec<Plan> {
 async fn run_plan(client: &Client, store: &AsyncProgramStore, plan: Plan, permits: &Semaphore) {
     let permit = permits.acquire().await;
     match plan {
-        Plan::Messages(room_id, fetch_id) => {
+        Plan::Messages(room_id, fetch_id, message_need) => {
             let limit = MIN_MSG_LOAD;
             let client = client.clone();
             let store_clone = store.clone();
 
             let res = load_older_one(&client, &room_id, fetch_id, limit).await;
             let mut locked = store.lock().await;
-            load_insert(room_id, res, locked.deref_mut(), store_clone);
+            load_insert(room_id, res, locked.deref_mut(), store_clone, message_need);
         },
         Plan::Members(room_id) => {
             let res = members_load(client, &room_id).await;
@@ -282,6 +278,9 @@ async fn load_older_one(
     limit: u32,
 ) -> MessageFetchResult {
     if let Some(room) = client.get_room(room_id) {
+        // Update cached encryption state. This is a noop if the state is already cached.
+        let _ = room.request_encryption_state().await;
+
         let mut opts = match &fetch_id {
             Some(id) => MessagesOptions::backward().from(id.as_str()),
             None => MessagesOptions::backward(),
@@ -293,10 +292,8 @@ async fn load_older_one(
         let mut msgs = vec![];
 
         for ev in chunk.into_iter() {
-            let msg = match ev.event.deserialize() {
-                Ok(AnyTimelineEvent::MessageLike(msg)) => msg,
-                Ok(AnyTimelineEvent::State(_)) => continue,
-                Err(_) => continue,
+            let Ok(msg) = ev.into_raw().deserialize() else {
+                continue;
             };
 
             let event_id = msg.event_id();
@@ -311,6 +308,7 @@ async fn load_older_one(
                 },
             };
 
+            let msg = msg.into_full_event(room_id.to_owned());
             msgs.push((msg, receipts));
         }
 
@@ -325,6 +323,7 @@ fn load_insert(
     res: MessageFetchResult,
     locked: &mut ProgramStore,
     store: AsyncProgramStore,
+    message_needs: Vec<MessageNeed>,
 ) {
     let ChatStore { presences, rooms, worker, picker, settings, .. } = &mut locked.application;
     let info = rooms.get_or_default(room_id.clone());
@@ -338,37 +337,57 @@ fn load_insert(
                 let _ = presences.get_or_default(sender);
 
                 for user_id in receipts {
-                    info.set_receipt(user_id, msg.event_id().to_owned());
+                    info.set_receipt(ReceiptThread::Main, user_id, msg.event_id().to_owned());
                 }
 
                 match msg {
-                    AnyMessageLikeEvent::RoomEncrypted(msg) => {
+                    AnyTimelineEvent::MessageLike(AnyMessageLikeEvent::RoomEncrypted(msg)) => {
                         info.insert_encrypted(msg);
                     },
-                    AnyMessageLikeEvent::RoomMessage(msg) => {
+                    AnyTimelineEvent::MessageLike(AnyMessageLikeEvent::RoomMessage(msg)) => {
                         info.insert_with_preview(
                             room_id.clone(),
                             store.clone(),
-                            *picker,
+                            picker.clone(),
                             msg,
                             settings,
                             client.media(),
                         );
                     },
-                    AnyMessageLikeEvent::Reaction(ev) => {
+                    AnyTimelineEvent::MessageLike(AnyMessageLikeEvent::Reaction(ev)) => {
                         info.insert_reaction(ev);
                     },
-                    _ => continue,
+                    AnyTimelineEvent::MessageLike(_) => {
+                        continue;
+                    },
+                    AnyTimelineEvent::State(msg) => {
+                        if settings.tunables.state_event_display {
+                            info.insert_any_state(msg.into());
+                        }
+                    },
                 }
             }
 
             info.fetch_id = fetch_id.map_or(RoomFetchStatus::Done, RoomFetchStatus::HaveMore);
+
+            // check if more are needed
+            let needs: Vec<_> = message_needs
+                .into_iter()
+                .filter(|need| !info.keys.contains_key(&need.event_id) && need.ttl > 0)
+                .map(|mut need| {
+                    need.ttl -= 1;
+                    need
+                })
+                .collect();
+            if !needs.is_empty() {
+                locked.application.need_load.need_messages_all(room_id, needs);
+            }
         },
         Err(e) => {
             warn!(room_id = room_id.as_str(), err = e.to_string(), "Failed to load older messages");
 
             // Wait and try again.
-            locked.application.need_load.insert(room_id, Need::MESSAGES);
+            locked.application.need_load.need_messages_all(room_id, message_needs);
         },
     }
 }
@@ -440,7 +459,7 @@ async fn refresh_rooms(client: &Client, store: &AsyncProgramStore) {
     let mut dms = vec![];
 
     for room in client.invited_rooms().into_iter() {
-        let name = room.display_name().await.unwrap_or(DisplayName::Empty).to_string();
+        let name = room.cached_display_name().unwrap_or(RoomDisplayName::Empty).to_string();
         let tags = room.tags().await.unwrap_or_default();
 
         names.push((room.room_id().to_owned(), name));
@@ -455,7 +474,7 @@ async fn refresh_rooms(client: &Client, store: &AsyncProgramStore) {
     }
 
     for room in client.joined_rooms().into_iter() {
-        let name = room.display_name().await.unwrap_or(DisplayName::Empty).to_string();
+        let name = room.cached_display_name().unwrap_or(RoomDisplayName::Empty).to_string();
         let tags = room.tags().await.unwrap_or_default();
 
         names.push((room.room_id().to_owned(), name));
@@ -490,31 +509,36 @@ async fn refresh_rooms_forever(client: &Client, store: &AsyncProgramStore) {
 
 async fn send_receipts_forever(client: &Client, store: &AsyncProgramStore) {
     let mut interval = tokio::time::interval(Duration::from_secs(2));
-    let mut sent = HashMap::<OwnedRoomId, OwnedEventId>::default();
+    let mut sent: HashMap<OwnedRoomId, HashMap<ReceiptThread, OwnedEventId>> = Default::default();
 
     loop {
         interval.tick().await;
 
-        let locked = store.lock().await;
-        let user_id = &locked.application.settings.profile.user_id;
-        let updates = client
-            .joined_rooms()
-            .into_iter()
-            .filter_map(|room| {
-                let room_id = room.room_id().to_owned();
-                let info = locked.application.rooms.get(&room_id)?;
-                let new_receipt = info.get_receipt(user_id)?;
-                let old_receipt = sent.get(&room_id);
-                if Some(new_receipt) != old_receipt {
-                    Some((room_id, new_receipt.clone()))
-                } else {
-                    None
+        let mut locked = store.lock().await;
+        let ChatStore { settings, open_notifications, rooms, .. } = &mut locked.application;
+        let user_id = &settings.profile.user_id;
+
+        let mut updates = Vec::new();
+        for room in client.joined_rooms() {
+            let room_id = room.room_id();
+            let Some(info) = rooms.get(room_id) else {
+                continue;
+            };
+
+            let changed = info.receipts(user_id).filter_map(|(thread, new_receipt)| {
+                let old_receipt = sent.get(room_id).and_then(|ts| ts.get(thread));
+                let changed = Some(new_receipt) != old_receipt;
+                if changed {
+                    open_notifications.remove(room_id);
                 }
-            })
-            .collect::<Vec<_>>();
+                changed.then(|| (room_id.to_owned(), thread.to_owned(), new_receipt.to_owned()))
+            });
+
+            updates.extend(changed);
+        }
         drop(locked);
 
-        for (room_id, new_receipt) in updates {
+        for (room_id, thread, new_receipt) in updates {
             use matrix_sdk::ruma::api::client::receipt::create_receipt::v3::ReceiptType;
 
             let Some(room) = client.get_room(&room_id) else {
@@ -522,15 +546,11 @@ async fn send_receipts_forever(client: &Client, store: &AsyncProgramStore) {
             };
 
             match room
-                .send_single_receipt(
-                    ReceiptType::Read,
-                    ReceiptThread::Unthreaded,
-                    new_receipt.clone(),
-                )
+                .send_single_receipt(ReceiptType::Read, thread.to_owned(), new_receipt.clone())
                 .await
             {
                 Ok(()) => {
-                    sent.insert(room_id, new_receipt);
+                    sent.entry(room_id).or_default().insert(thread, new_receipt);
                 },
                 Err(e) => tracing::warn!(?room_id, "Failed to set read receipt: {e}"),
             }
@@ -549,7 +569,7 @@ pub async fn do_first_sync(client: &Client, store: &AsyncProgramStore) -> Result
     let mut filter = FilterDefinition::default();
     filter.room = room_ev;
 
-    let settings = SyncSettings::new().filter(filter.into());
+    let settings = SyncSettings::new().filter(filter.into()).timeout(Duration::from_secs(0));
 
     client.sync_once(settings).await?;
 
@@ -562,12 +582,12 @@ pub async fn do_first_sync(client: &Client, store: &AsyncProgramStore) -> Result
 
     for room in sync_info.rooms.iter() {
         let room_id = room.as_ref().0.room_id().to_owned();
-        need_load.insert(room_id, Need::MESSAGES);
+        need_load.need_messages(room_id);
     }
 
     for room in sync_info.dms.iter() {
         let room_id = room.as_ref().0.room_id().to_owned();
-        need_load.insert(room_id, Need::MESSAGES);
+        need_load.need_messages(room_id);
     }
 
     Ok(())
@@ -603,7 +623,7 @@ fn oneshot<T>() -> (ClientReply<T>, ClientResponse<T>) {
     return (reply, response);
 }
 
-pub type FetchedRoom = (MatrixRoom, DisplayName, Option<Tags>);
+pub type FetchedRoom = (MatrixRoom, RoomDisplayName, Option<Tags>);
 
 pub enum WorkerTask {
     Init(AsyncProgramStore, ClientReply<()>),
@@ -700,7 +720,7 @@ async fn create_client_inner(
         .build()
         .unwrap();
 
-    let req_config = RequestConfig::new().timeout(req_timeout).retry_timeout(req_timeout);
+    let req_config = RequestConfig::new().timeout(req_timeout).max_retry_time(req_timeout);
 
     // Set up the Matrix client for the selected profile.
     let builder = Client::builder()
@@ -1001,7 +1021,7 @@ impl ClientWorker {
                     info.insert_with_preview(
                         room_id.to_owned(),
                         store.clone(),
-                        *picker,
+                        picker.clone(),
                         full_ev,
                         settings,
                         client.media(),
@@ -1043,14 +1063,32 @@ impl ClientWorker {
                         let Some(receipts) = receipts.get(&ReceiptType::Read) else {
                             continue;
                         };
-                        for user_id in receipts.keys() {
-                            info.set_receipt(user_id.to_owned(), event_id.clone());
+                        for (user_id, rcpt) in receipts.iter() {
+                            info.set_receipt(
+                                rcpt.thread.clone(),
+                                user_id.to_owned(),
+                                event_id.clone(),
+                            );
                         }
                     }
                 }
             },
         );
 
+        if self.settings.tunables.state_event_display {
+            let _ = self.client.add_event_handler(
+                |ev: AnySyncStateEvent, room: MatrixRoom, store: Ctx<AsyncProgramStore>| {
+                    async move {
+                        let room_id = room.room_id();
+                        let mut locked = store.lock().await;
+
+                        let info = locked.application.get_room_info(room_id.to_owned());
+                        info.insert_any_state(ev);
+                    }
+                },
+            );
+        }
+
         let _ = self.client.add_event_handler(
             |ev: OriginalSyncRoomRedactionEvent,
              room: MatrixRoom,
@@ -1058,11 +1096,15 @@ impl ClientWorker {
                 async move {
                     let room_id = room.room_id();
                     let room_info = room.clone_info();
-                    let room_version = room_info.room_version().unwrap_or(&RoomVersionId::V1);
+                    let rules = &room_info
+                        .room_version()
+                        .and_then(RoomVersionId::rules)
+                        .unwrap_or(RoomVersionId::V1.rules().unwrap())
+                        .redaction;
 
                     let mut locked = store.lock().await;
                     let info = locked.application.get_room_info(room_id.to_owned());
-                    info.redact(ev, room_version);
+                    info.redact(ev, rules);
                 }
             },
         );
@@ -1076,11 +1118,12 @@ impl ClientWorker {
                     let room_id = room.room_id();
                     let user_id = ev.state_key;
 
-                    let ambiguous_name =
-                        ev.content.displayname.as_deref().unwrap_or_else(|| user_id.localpart());
+                    let ambiguous_name = DisplayName::new(
+                        ev.content.displayname.as_deref().unwrap_or_else(|| user_id.as_str()),
+                    );
                     let ambiguous = client
-                        .store()
-                        .get_users_with_display_name(room_id, ambiguous_name)
+                        .state_store()
+                        .get_users_with_display_name(room_id, &ambiguous_name)
                         .await
                         .map(|users| users.len() > 1)
                         .unwrap_or_default();
@@ -1217,7 +1260,7 @@ impl ClientWorker {
             let settings = self.settings.clone();
 
             async move {
-                while !client.logged_in() {
+                while !client.is_active() {
                     tokio::time::sleep(Duration::from_millis(100)).await;
                 }
 
@@ -1309,7 +1352,7 @@ impl ClientWorker {
         // Remove the session.json file.
         std::fs::remove_file(&self.settings.session_json)?;
 
-        Ok(Some(InfoMessage::from("Sucessfully logged out")))
+        Ok(Some(InfoMessage::from("Successfully logged out")))
     }
 
     async fn direct_message(&mut self, user: OwnedUserId) -> IambResult<OwnedRoomId> {
@@ -1346,7 +1389,7 @@ impl ClientWorker {
 
     async fn get_room(&mut self, room_id: OwnedRoomId) -> IambResult<FetchedRoom> {
         if let Some(room) = self.client.get_room(&room_id) {
-            let name = room.display_name().await.map_err(IambError::from)?;
+            let name = room.cached_display_name().ok_or_else(|| IambError::UnknownRoom(room_id))?;
             let tags = room.tags().await.map_err(IambError::from)?;
 
             Ok((room, name, tags))
@@ -1389,9 +1432,9 @@ impl ClientWorker {
         req.limit = Some(1000u32.into());
         req.max_depth = Some(1u32.into());
 
-        let resp = self.client.send(req, None).await.map_err(IambError::from)?;
+        let resp = self.client.send(req).await.map_err(IambError::from)?;
 
-        let rooms = resp.rooms.into_iter().map(|chunk| chunk.room_id).collect();
+        let rooms = resp.rooms.into_iter().map(|chunk| chunk.summary.room_id).collect();
 
         Ok(rooms)
     }
